/*
 *  linux/kernel/tty_io.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include "tty_ed.h"
#include "softspace.h"
#include <linux/timer.h>
#include <linux/ctype.h>
#include <linux/kd.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/keyboard.h>
#include <linux/interrupt.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/bitops.h>

#include "use_count.h"	/* should be <linux/use_count.h> */

#define INPUT_ESC 0	/* introduces special codes in the secondary queue */
#define LITERAL_ESC 1	/* a literal INPUT_ESC character */
#define END_REC 2	/* END of record */


static void crted_init(struct tty_struct *);
static void crted_cleanup(struct tty_struct *);
static void crted_start_write(struct tty_struct *);
static void crted_process_input(struct tty_struct *);
static int crted_read_chan(struct tty_struct *, struct file *, char *, int);
static int crted_write_chan(struct tty_struct *, struct file *, char *, int);
static int crted_select(struct inode *, struct file *, int, select_table *);
int crted_ioctl(struct inode *, struct file *, unsigned, unsigned);
void crted_flush_input(struct tty_struct *);
void crted_flush_output(struct tty_struct *);
void crted_copy_to_cooked(struct tty_struct * tty);
static void crted_tick(struct tty_struct *);

struct ldisc crted_ldisc = {
	8, crted_init,
	{crted_cleanup, crted_start_write,
	 crted_read_chan, crted_write_chan, crted_select, crted_ioctl,
	 crted_flush_input, crted_flush_output},
	{crted_process_input, tty_std_hangup, crted_tick}
};
	
void tty_in_char(int, struct tty_struct *);
void tty_echo(int, struct tty_struct *);
void tty_erase_chars(char *, struct tty_struct *);
void tty_reprint(struct tty_struct *);
static int crted_insize(struct tty_struct *);
void sendsig(int, struct tty_struct *);
static int set_termios(struct tty_struct *, unsigned, unsigned, int);
static void tty_set_ldisc(struct termios *, struct tty_struct *);
static void crted_canon_off(struct tty_struct *);
static void crted_canon_on(struct tty_struct *);

char softint_space[SOFTINT_SPACE];	/* for now */


#ifdef SEPARATE_MODULE
/* Module initialization routine */
int init() {
	printk("sizeof tty_struct = %d\n", sizeof (struct tty_struct));
	if (tty_add_ldisc(&crted_ldisc) < 0)
		return -1;
	return 0;
}

void cleanup() {
	tty_delete_ldisc(&crted_ldisc);
}
#endif

/*
 * Called when we switch to mode 2.
 * To do:  restructure to handle running out of memory.
 */
static void crted_init(struct tty_struct *tty)
{
#ifdef SEPARATE_MODULE
	INC_USE_COUNT;
#endif
	printk("crted_init called\n");
	memset(&tty->curline, 0, 4096 - offsetof(struct tty_struct, curline));
	tty->endcurline = tty->curline;		/* NOTE: may discard text */
	tty->endhist = 0L;
	tty->write_ipos = tty->curline;
}

static void crted_cleanup(struct tty_struct *tty)
{
	printk("crted_cleanup called\n");
	edclose(tty);
	if (tty->histarea) {
		free_page((unsigned long) tty->histarea);
		tty->histarea = NULL;
	}
#ifdef SEPARATE_MODULE
	DEC_USE_COUNT;
#endif
}


/*TEMPORARY*/
int my_get_tty_queue(struct tty_queue * queue)
{
	int result = -1;
	unsigned long flags;

	/* printk("<getq 0x%X 0x%X>", ((unsigned *) &queue)[-1], queue); */
	__asm__ __volatile__("pushfl ; popl %0 ; cli":"=r" (flags));
	if (queue->tail != queue->head) {
		result = queue->buf[queue->tail];
		queue->tail = (queue->tail + 1) & (TTY_BUF_SIZE-1);
	}
	__asm__ __volatile__("pushl %0 ; popfl"::"r" (flags));
	return result;
}


#if 0
/*
 * Special version of sleep that turns on software interrupts during
 * the sleep.
 */
void my_sleep_on(struct wait_queue **p)
{
	unsigned long flags;
	unsigned long mask;
	struct wait_queue wait = { current, NULL };

	if (!p)
		return;
	if (current == task[0])
		panic("task[0] trying to sleep");
	current->state = TASK_INTERRUPTIBLE;
	add_wait_queue(p, &wait);
	save_flags(flags);
	mask = bh_mask;
	bh_mask = 0xFFFFFFFF;
	sti();
	if (bh_active) {
		/* later change to redo_bottom_half */
		cli();
		do_bottom_half();
		sti();
	}
	schedule();
	remove_wait_queue(p, &wait);
	bh_mask = mask;
	restore_flags(flags);
}
#endif



static void crted_start_write(struct tty_struct * tty)
{

	if (!tty->write || EMPTY(&tty->write_q))
		return;
	SOFT_INT_OFF;		/* must be nestable */
	if (set_bit(TTY_WRITE_BUSY,&tty->flags))
		printk("crted_start_write: bit already set\n");
	tty->write(tty);
	if (clear_bit(TTY_WRITE_BUSY,&tty->flags))
		printk("crted_start_write: bit already cleared\n");
	SOFT_INT_ON;
}

void crted_process_input(struct tty_struct * tty)
{
	if (!tty || EMPTY(&tty->read_q))
		return;
	if (set_bit(TTY_READ_BUSY, &tty->flags))
		printk("crted_process_input: bit already set\n");
	crted_copy_to_cooked(tty);
	if (clear_bit(TTY_READ_BUSY, &tty->flags))
		printk("crted_process_input: bit already cleared\n");
}


void crted_copy_to_cooked(struct tty_struct * tty)
{
	int c;

	if (!tty->write) {
		printk("copy_to_cooked: tty %d has null write routine\n",
		       tty->line);
	}
	while (1) {
		c = my_get_tty_queue(&tty->read_q);
		if (c < 0)
			break;
		if (! C_READ(tty))
			continue;
		if (c &~ 0xFF)  printk("<ctc %X>", c);		/*DEBUG*/
		c &=~ TTY_OE;
		if (! I_INPCK(tty))
			c &=~ TTY_PE;
		if (c & (TTY_PE | TTY_FE)) {
			if (I_IGNPAR(tty))
				continue;
		}
		if (c == TTY_BREAK_CHAR) {
			if (I_IGNBRK(tty))
				continue;
			if (I_BRKINT(tty)) {
				sendsig(SIGINT, tty);
				crted_flush_input(tty);
				crted_flush_output(tty);
				continue;
			}
		}
		if (I_STRP(tty))
			c &=~ 0x80;
		if (I_IXON(tty)) {
			if (I_IXANY(tty) && L_IEXTEN(tty) && tty->stopped) {
				tty->status_changed = 1;
				tty->ctrl_status |= TIOCPKT_START;
				tty->stopped = 0;
				continue;
			}
			if ((STOP_CHAR(tty) != __DISABLED_CHAR) &&
			    (c==STOP_CHAR(tty))) {
			        tty->status_changed = 1;
				tty->ctrl_status |= TIOCPKT_STOP;
				tty->stopped=1;
				continue;
			}
			if ((START_CHAR(tty) != __DISABLED_CHAR) &&
			     (c==START_CHAR(tty))) {
			        tty->status_changed = 1;
				tty->ctrl_status |= TIOCPKT_START;
				tty->stopped=0;
				continue;
			}
		}
		if (L_IEXTEN(tty) && L_CANON(tty) && tty->keytabs)
			inchar(c & 0xFF, tty);		/* rename */
		else {
			if (I_PARMRK(tty)) {
				if (c == 0377) {
					/* This assumes that 0377 is not */
					/* a special character (which will */
					/* be the case if _POSIX_VDISABLE */
					/* is 0377. */
					tty_in_char(0377, tty);
				}
				if (c &~ 0xFF) {
					tty->lnext = 1;
					tty_in_char(0377, tty);
					tty->lnext = 1;
					tty_in_char(0, tty);
					tty->lnext = 1;
				}
			} else if (c &~ 0xFF) {
				c = 0;
				tty->lnext = 1;
			}
			tty_in_char(c & 0xFF, tty);
		}
	}
	if (I_IXOFF(tty)
	 && ! test_bit(TTY_SQ_THROTTLED, &tty->flags)
	 && ! EMPTY(&tty->secondary)
	 && crted_insize(tty) > 32) {
		if (tty->throttle)
			tty->throttle(tty, TTY_THROTTLE_SQ_FULL);
		else
			put_tty_queue(STOP_CHAR(tty), &tty->write_q);
		set_bit(TTY_SQ_THROTTLED, &tty->flags);
	}
	if (L_IEXTEN(tty) && L_CANON(tty) && tty->keytabs)
		refresh(tty);
	TTY_WRITE_FLUSH(tty);
	if (!EMPTY(&tty->secondary))
		wake_up_interruptible(&tty->secondary.proc_list);
	if (tty->write_q.proc_list && LEFT(&tty->write_q) > TTY_BUF_SIZE/2)
		wake_up_interruptible(&tty->write_q.proc_list);
}

void tty_in_char(int c, struct tty_struct *tty)
{
	char *p;

	if (tty->lnext) {
		tty->lnext = 0;
	} else if (c != __DISABLED_CHAR) {
		if (c=='\r') {
			if (I_NOCR(tty))
				return;
			if (I_CRNL(tty))
				c = '\n';
		} else if (c=='\n') {
			if (I_NLCR(tty))
				c = '\r';
		}
#ifdef notdef
		if (I_UCLC(tty) && L_IEXTEN(tty))
			c=tolower(c);
#endif
		if (L_ISIG(tty) && ! tty->lnext) {
			if (c==INTR_CHAR(tty)) {
				sendsig(SIGINT, tty);
				return;
			}
			if (c==QUIT_CHAR(tty)) {
				sendsig(SIGQUIT, tty);
				return;
			}
			if (c==SUSPEND_CHAR(tty)) {
				if (!is_orphaned_pgrp(tty->pgrp))
					sendsig(SIGTSTP, tty);
				return;
			}
		}
		if (c == LNEXT_CHAR(tty) && L_IEXTEN(tty)) {
			tty->lnext = 1;
			return;
		}
		if (L_CANON(tty)) {
			if (c==KILL_CHAR(tty)) {
				p = tty->endcurline;
				tty->endcurline = tty->curline;
				if (L_ECHOK(tty))
					tty_erase_chars(p, tty);
				else
					tty_echo(c, tty);
				return;
			}
			if (c==ERASE_CHAR(tty)) {
				p = tty->endcurline;
				if (p != tty->curline) {
					tty->endcurline = p - 1;
					if (L_ECHOE(tty))
						tty_erase_chars(p, tty);
				}
				if (! L_ECHOE(tty))
					tty_echo(c, tty);
				return;
			}
			if (c == WERASE_CHAR(tty) && L_IEXTEN(tty)) {
				char *saveend;
				p = tty->endcurline;
				while (p > tty->curline
				 && (p[-1] == ' ' || p[-1] == '\t'))
					p--;
				while (p > tty->curline
				 && (p[-1] != ' ' && p[-1] != '\t'))
					p--;
				saveend = tty->endcurline;
				tty->endcurline = p;
				if (L_ECHOE(tty))
					tty_erase_chars(saveend, tty);
				else
					tty_echo(c, tty);
				return;
			}
			if (c == REPRINT_CHAR(tty) && L_IEXTEN(tty)) {
				tty_reprint(tty);
				return;
			}
			if (c=='\n' || c==EOF_CHAR(tty)) {
				if (LEFT(&tty->secondary) <
				     tty->endcurline - tty->curline + 3) {
					/* fail */
					return;
				}
				for (p = tty->curline ; p < tty->endcurline ; p++) {
					put_tty_queue(*p, &tty->secondary);
					if (*p == INPUT_ESC)
						put_tty_queue(LITERAL_ESC,
								&tty->secondary);
				}
				if (c == '\n')
					put_tty_queue('\n', &tty->secondary);
				put_tty_queue(INPUT_ESC, &tty->secondary);
				put_tty_queue(END_REC, &tty->secondary);
				wake_up_interruptible(&tty->secondary.proc_list);
				tty->endcurline = tty->curline;
				tty->writecol = tty->column;
				tty->write_ipos = tty->curline;
				goto echo;
			}
		}
	}
	if (L_CANON(tty)) {
		if (tty->endcurline < tty->curline + MAXLLEN) {
			*tty->endcurline++ = c;
		}
	} else {
#if 0
		if (LEFT(&tty->secondary) >= 2) {
			put_tty_queue(c, &tty->secondary);
			if (c == INPUT_ESC)
				put_tty_queue(LITERAL_ESC, &tty->secondary);
			wake_up_interruptible(&tty->secondary.proc_list);
		}
#else
		put_tty_queue(c, &tty->secondary);
		tty->readtimeout = tty->chartimeout;
		if (CHARS(&tty->secondary) >= tty->minread)
			wake_up_interruptible(&tty->secondary.proc_list);
#endif
	}
echo:
	if (L_ECHO(tty)) {
		tty_echo(c, tty);
	} else if (L_ECHONL(tty) && L_CANON(tty)) {
		/* Should we output a CR if ONLCR is set? */
		/* Normally the CR is not needed. */
		if (c == '\n')
			put_tty_queue('\n', &tty->write_q);
	}
}


void tty_echo(int c, struct tty_struct *tty) {
	if (! L_ECHO(tty))
		return;
	if (L_ECHOCTL(tty) && L_IEXTEN(tty)
	 && ((c < 32 && c != '\t' && c != '\n') || c == 127)) {
		tty_put_char('^', tty);
		tty_put_char(c == 127? '?' : c + 64, tty);
	} else {
		tty_put_char(c, tty);
	}
}


void tty_erase_chars(char *oldend, struct tty_struct *tty)
{
	register char *p;
	int col;
	int begincol;
	int reprint;
	int ncols;

	p = tty->write_ipos;
	col = tty->writecol;
	begincol = col;
	reprint = 1;
	for (;;) {
		if (p == tty->endcurline) {
			begincol = col;
			reprint = 0;
		}
		if (p == oldend) {
			ncols = col - begincol;
			break;
		}
		if (*p == '\t') {
			col = col + 8 &~ 07;
		} else if (*p < 32 || *p == 127) {
			if (L_ECHOCTL(tty))
				col += 2;
		} else {
			col++;
		}
		p++;
	}
	while (--ncols >= 0) {
		tty_put_char('\b', tty);
		tty_put_char(' ', tty);
		tty_put_char('\b', tty);
	}
	if (reprint)
		tty_reprint(tty);
}


void tty_reprint(struct tty_struct *tty)
{
	register char *p;

	tty_put_char('\n', tty);
	tty->write_ipos = tty->curline;
	tty->writecol = 0;
	for (p = tty->curline ; p < tty->endcurline ; p++)
		tty_put_char(*p, tty);
}


static int crted_insize(struct tty_struct *tty)
{
	int size;
	char *p;

	size = CHARS(&tty->secondary);
	if (L_CANON(tty)) {
		size += tty->endcurline - tty->curline;
		for (p = tty->curline ; p < tty->endcurline ; p++) {
			if (*p == INPUT_ESC)
				size++;
		}
	}
	return size;
}


/*
 * Read into user space.  To avoid race conditions, we place the data in
 * the tty structure, and then do a memcpy_tofs at the end of the routine.
 */
static int crted_read_chan(struct tty_struct * tty, struct file * file, char * buf, int nr)
{
	struct wait_queue wait = { current, NULL };
	int c;
	char *b;
	int minimum, time;
	int rc;

	while (set_bit(TTY_READING, &tty->flags)) {
		set_bit(TTY_WAIT_READ, &tty->flags);
		interruptible_sleep_on(&tty->secondary.proc_list);
		if (current->signal &~ current->blocked)
			return -ERESTARTSYS;
	}
	if (tty->readtimeout != 0) {
		printk("crted_read_chan:  readtimeout != 0\n");
		tty->readtimeout = 0;
	}
	if (tty->gottimeout != 0) {
		printk("crted_read_chan:  gottimeout != 0\n");
		tty->gottimeout = 0;
	}
	if (file->f_flags & O_NONBLOCK) {
		minimum = 0;
		time = 0;
	} else if (L_CANON(tty)) {
		minimum = 1;
		time = 0;
	} else {
		minimum = tty->termios->c_cc[VMIN];
		time = tty->termios->c_cc[VTIME];
	}
	if (time) {
		if (minimum == 0) {
			minimum = 1;
			tty->readtimeout = time + 1;
		} else {
			tty->chartimeout = time + 1;
			if (! EMPTY(&tty->secondary))
				tty->readtimeout = time + 1;
		}
	}
	tty->minread = minimum;
	if (nr > 256)
		nr = 256;
	rc = 1;		/* means don't return yet */
	SOFT_INT_OFF;
	while (CHARS(&tty->secondary) < minimum && ! tty->gottimeout) {
		add_wait_queue(&tty->secondary.proc_list, &wait);
		current->state = TASK_INTERRUPTIBLE;
		if (hung_up(file))
			rc = 0;
		else if (current->signal & ~current->blocked) 
			rc = -ERESTARTSYS;
		else {
			SOFT_INT_ON;
			schedule();
			SOFT_INT_OFF;
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&tty->secondary.proc_list, &wait);
		if (rc <= 0) {
			SOFT_INT_ON;
			goto out;
		}
	}
	b = tty->readbuf;
	for (;;) {
		if (--nr < 0 && (CHARS(&tty->secondary) < 2 ||
				 FIRST(&tty->secondary) != INPUT_ESC ||
				 SECOND(&tty->secondary) != END_REC))
			break;
		if ((c = my_get_tty_queue(&tty->secondary)) < 0)
			break;
		if (c == INPUT_ESC && L_CANON(tty)) {
			switch (my_get_tty_queue(&tty->secondary)) {
				case END_REC:
					goto break_loop;
				default:	/* should be LITERAL_ESC */
					break;
			}
		}
		*b++ = c;
	};
break_loop:
	if (test_bit(TTY_SQ_THROTTLED, &tty->flags)
	 && (EMPTY(&tty->secondary) || crted_insize(tty) < 16)) {
		if (tty->throttle)
			tty->throttle(tty, TTY_THROTTLE_SQ_AVAIL);
		else
			put_tty_queue(START_CHAR(tty), &tty->write_q);
		clear_bit(TTY_SQ_THROTTLED, &tty->flags);
		TTY_WRITE_FLUSH(tty);
	}
	SOFT_INT_ON;
	rc = b - tty->readbuf;
	verify_area(buf, rc);
	memcpy_tofs(buf, tty->readbuf, rc);
out:
	if (! L_CANON(tty)) {
		tty->chartimeout = 0;
		tty->readtimeout = 0;
		tty->gottimeout = 0;
		tty->minread = 32767;		/* avoid unnecessary wakeups */
	}
	clear_bit(TTY_READING, &tty->flags);
	if (! clear_bit(TTY_WAIT_READ, &tty->flags))
		wake_up_interruptible(&tty->secondary.proc_list);
	return rc;
}


static int crted_write_chan(struct tty_struct * tty, struct file * file, char * buf, int nr)
{
	struct wait_queue wait = { current, NULL };
	char *b;
	char c;

	if (nr == 0)
		return 0;
	if (nr < 0)
		return -EINVAL;
	if (tty->write == NULL) {
		printk("Missing tty write routine\n");
		return -ENXIO;
	}
	while (set_bit(TTY_WRITING, &tty->flags)) {
		set_bit(TTY_WAIT_WRITE, &tty->flags);
		interruptible_sleep_on(&tty->write_q.proc_list);
		if (current->signal &~ current->blocked)
			return -ERESTARTSYS;
	}
	SOFT_INT_OFF;
	if (tty->keytabs && tty->eddisplay)
		dispoff(tty->column, tty);
	SOFT_INT_ON;
	b = buf;
	for (;;) {
		if (nr <= 0)
			break;
		if (current->signal & ~current->blocked)
			break;
		if (hung_up(file))
			break;
		for (;;) {
			if (nr <= 0) {
				tty_write_flush(tty);
				goto break2;
			}
			if (LEFT(&tty->write_q) <= 10) {
				tty_write_flush(tty);
				if (LEFT(&tty->write_q) <= 10)
					break;
			}
			c = get_fs_byte(b++);
			nr--;
#if 0	/* for testing only */
			put_tty_queue(c, &tty->write_q);
#else
			SOFT_INT_OFF;
			tty_put_char(c, tty);
			tty->writecol = tty->column;
			tty->write_ipos = tty->endcurline;
			SOFT_INT_ON;
#endif
		}
		if (file->f_flags & O_NONBLOCK) {
			break;
		}
		add_wait_queue(&tty->write_q.proc_list, &wait);
		current->state = TASK_INTERRUPTIBLE;
		if (! EMPTY(&tty->write_q))
			schedule();
		current->state = TASK_RUNNING;
		remove_wait_queue(&tty->write_q.proc_list, &wait);
	}
break2:
	nr = b - buf;
	clear_bit(TTY_WRITING, &tty->flags);
	if (! clear_bit(TTY_WAIT_WRITE, &tty->flags))
		wake_up_interruptible(&tty->write_q.proc_list);
	if (nr == 0 && current->signal &~ current->blocked)
		return -ERESTARTSYS;
	return nr;
}

static void crted_tick(struct tty_struct *tty)
{

	if (tty->eddisplay == 0 && (tty->flags & 1<<TTY_WRITING) == 0
	 && tty->keytabs && L_CANON(tty) && L_IEXTEN(tty)) {
		refresh(tty);
		TTY_WRITE_FLUSH(tty);
	}
	if (tty->readtimeout && --tty->readtimeout == 0) {
		tty->gottimeout = 1;
		wake_up_interruptible(&tty->secondary.proc_list);
	}
}


static int crted_select(struct inode * inode, struct file * filp, int sel_type, select_table * wait)
{
	int dev;
	struct tty_struct * tty;

	dev = filp->f_rdev;
	if (MAJOR(dev) != 4) {
		printk("tty_select: tty pseudo-major != 4\n");
		return 0;
	}
	dev = MINOR(filp->f_rdev);
	tty = TTY_TABLE(dev);
	if (!tty) {
		printk("tty_select: tty struct for dev %d was NULL\n", dev);
		return 0;
	}
	switch (sel_type) {
		case SEL_IN:
			if (!EMPTY(&tty->secondary))
				return 1;
			/* following MAY be needed for pty slaves */
			if (tty->link && !tty->link->count)
				return 1;

			select_wait(&tty->secondary.proc_list, wait);
			return 0;
		case SEL_OUT:
			if (LEFT(&tty->write_q) > 10)
				return 1;
			select_wait(&tty->write_q.proc_list, wait);
			return 0;
		case SEL_EX:
			return 0;
	}
	return 0;
}


void ttyoutc(int c, struct tty_struct *tty)
{
	put_tty_queue(c, &tty->write_q);
}

void outreal(char *p, int size, struct tty_struct *tty)
{
	while (--size >= 0)
		put_tty_queue(*p++, &tty->write_q);
}

/* LATER: return status */
void senddata(char *p, int size, struct tty_struct *tty)
{
	char lastc;

	if (LEFT(&tty->secondary) < size + 2)
		return;
	while (--size >= 0) {
		lastc = *p++;
		put_tty_queue(lastc, &tty->secondary);
	}
	put_tty_queue(INPUT_ESC, &tty->secondary);
	put_tty_queue(END_REC, &tty->secondary);
}

void sendsig(int signo, struct tty_struct *tty)
{
	kill_pg(tty->pgrp, signo, 1);
	if (! (tty->termios->c_lflag & NOFLSH)) {
		crted_flush_input(tty);
		crted_flush_output(tty);
	}
}


static void flush_queue(struct tty_queue * queue)
{
	cli();
	queue->head = queue->tail;
	sti();
}

void crted_flush_input(struct tty_struct * tty)
{
        tty->status_changed = 1;
	tty->ctrl_status |= TIOCPKT_FLUSHREAD;
#ifdef notdef	/* INCORRECT! */
	cli();
	tty->read_q.head = tty->read_q.tail;
	sti();
#endif
	tty->secondary.head = tty->secondary.tail;
/*
	if (tty->throttle && !clear_bit(TTY_SQ_THROTTLED, &tty->flags))
		tty->throttle(tty, TTY_THROTTLE_SQ_AVAIL);
*/
	wake_up_interruptible(&tty->read_q.proc_list);

	tty->endcurline = tty->curline;
	if (tty->keytabs)
		edflush(tty);

	if (test_bit(TTY_SQ_THROTTLED, &tty->flags)) {
		if (tty->throttle)
			tty->throttle(tty, TTY_THROTTLE_SQ_AVAIL);
		else
			put_tty_queue(START_CHAR(tty), &tty->write_q);
		clear_bit(TTY_SQ_THROTTLED, &tty->flags);
		TTY_WRITE_FLUSH(tty);
	}

	if ((tty = tty->link) != NULL) {
		flush_queue(&tty->write_q);
		wake_up_interruptible(&tty->write_q.proc_list);
	}
}

void crted_flush_output(struct tty_struct * tty)
{
   	tty->status_changed = 1;
	tty->ctrl_status |= TIOCPKT_FLUSHWRITE;
	cli();
	tty->write_q.head = tty->write_q.tail;
	sti();
	wake_up_interruptible(&tty->write_q.proc_list);
	if ((tty = tty->link) != NULL) {
		flush_queue(&tty->read_q);
		wake_up_interruptible(&tty->read_q.proc_list);
		flush_queue(&tty->secondary);
		tty->secondary.data = 0;
	}
}

int crted_ioctl(struct inode * inode, struct file * file,
	unsigned int cmd, unsigned int arg)
{
	int dev;
	struct tty_struct *tty;

	dev = MINOR(file->f_rdev);
	tty = TTY_TABLE(dev);
	if (tty == NULL) {
		printk("crted_ioctl: tty == NULL!\n");
		return -ENXIO;
	}
	switch (cmd) {
		case TCSETS:  case TCSETSW:  case TCSETSF:
		case TCSETA:  case TCSETAW:  case TCSETAF:
			return set_termios(tty, cmd, arg, dev);
		case 0x5426:		/* TIOCINITEDIT */
		{
			char *bufp;

			printk("TIOCINITEDIT called\n");
			if (arg) {
				if (tty->histarea == NULL) {
					bufp = (char *)get_free_page(GFP_KERNEL);
					tty->endhist = 0;
					tty->histarea = bufp;
				}
				edinit((char *) arg, tty);
			} else {
				edclose(tty);
			}
			return 0;
		}
		default:
			return -ENOSYS;
	}
}


static int set_termios(struct tty_struct *tty, unsigned cmd, unsigned arg, int channel)
{
	int actions;
	int format;
	struct termios new;
	union {
		struct termio termio;
	} u;

	cmd -= 2;
	actions = cmd & 03;
	format = cmd &~ 03;

	/* See if this operation is allowed. */
	/* If we try to set the state of terminal and we're not in the
	   foreground, send a SIGTTOU.  If the signal is blocked or
	   ignored, go ahead and perform the operation.  POSIX 7.2) */
	if (current->tty != channel
	 && tty->pgrp > 0 && tty->pgrp != current->pgrp) {
		if (is_orphaned_pgrp(current->pgrp))
			return -EIO;
		if (! is_ignored(SIGTTOU)) {
			kill_pg(current->pgrp,SIGTTOU,1);
			return -ERESTARTSYS;
		}
	}

	/* Copy the modes from user space. */
	switch (format) {
	case TCSETS - 2:
		memcpy_fromfs(&new, (void *) arg, sizeof new);
		break;
	case TCSETA - 2:
		memcpy_fromfs(&u.termio, (void *)arg, sizeof u.termio);
		new = *tty->termios;
		*(unsigned short *)&new.c_iflag = u.termio.c_iflag;
		*(unsigned short *)&new.c_oflag = u.termio.c_oflag;
		*(unsigned short *)&new.c_cflag = u.termio.c_cflag;
		*(unsigned short *)&new.c_lflag = u.termio.c_lflag;
		new.c_line = u.termio.c_line;
		memcpy(new.c_cc, u.termio.c_cc, NCC);
		break;
	default:
		printk("set_termios error: format 0x%X\n", format);
		return -EINVAL;
	}

	/* Perform any actions required, and disable soft-ints. */
	for (;;) {
		SOFT_INT_OFF;
		if (actions == TCSANOW || EMPTY(&tty->write_q))
			break;
		SOFT_INT_ON;
		wait_until_sent(tty);
		if (current->signal & ~current->blocked)
			return -ERESTARTSYS;
	}
	if (actions == TCSAFLUSH)
		flush_input(tty);

	/* Now set the modes. */
	if (L_CANON(tty)) {
		if ((new.c_lflag & ICANON) == 0)
			crted_canon_off(tty);
	} else {
		if ((new.c_lflag & ICANON) != 0)
			crted_canon_on(tty);
	}
	if (new.c_line != tty->termios->c_line)
		tty_set_ldisc(&new, tty);
	if (tty->dev.new_termios)
		tty->dev.new_termios(tty, &new);
	*tty->termios = new;
	SOFT_INT_ON;
	return 0;
}

/* Set the discipline of a tty line. */
static void tty_set_ldisc(struct termios *new, struct tty_struct *tty)
{
	register unsigned long ldisc = new->c_line;
	register struct ldisc *ldp, **ldpp;

	for (ldpp = ldiscs ; ; ldpp++) {
		if (ldpp >= &ldiscs[NR_LDISCS]) {
			new->c_line = tty->termios->c_line;
			return;
		}
		ldp = *ldpp;
		if (ldp != NULL && ldp->name == ldisc)
			break;
	}

	/* Shut down the current discipline. */
	tty->lb.process_input = tty_no_op;
	tty->lb.hangup = tty_std_hangup;
	tty->lb.timer = NULL;
	if (tty->la.close_ldisc)
		tty->la.close_ldisc(tty);

	/* Now set up the new line discipline. */
	tty->la = ldp->ldisc_above;
	if (ldp->open_ldisc)
		ldp->open_ldisc(tty);		/* must not block */
	tty->lb = ldp->ldisc_below;
	if (! EMPTY(&tty->read_q))
		tty->lb.process_input(tty);
}

static void crted_canon_off(struct tty_struct *tty)
{
	char *p;
	int c;
	int n;

	p = softint_space;
	while ((c = get_tty_queue(&tty->secondary)) >= 0) {
		if (c == INPUT_ESC) {
			c = get_tty_queue(&tty->secondary);
			if (c != END_REC)
				*p++ = INPUT_ESC;
		} else {
			*p++ = c;
		}
	}
	n = p - softint_space;
	p = softint_space;
	while (--n >= 0) {
		put_tty_queue(*p++, &tty->secondary);
	}
	p = tty->curline;
	while (p < tty->endcurline) {
		put_tty_queue(*p++, &tty->secondary);
	}
	tty->endcurline = tty->curline;
}


static void crted_canon_on(struct tty_struct *tty)
{
	char *p;
	int c;
	int n;
	char *lastline;

	p = softint_space;
	lastline = p;
	while ((c = get_tty_queue(&tty->secondary)) >= 0) {
		*p++ = c;
		if (c == '\n')
			lastline = p;
	}
	n = p - lastline;
	if (n > MAXLLEN)
		n = MAXLLEN;
	memcpy(tty->curline, lastline, n);
	tty->endcurline = tty->curline + n;
	p = softint_space;
	while (p < lastline) {
		c = *p++;
		put_tty_queue(c, &tty->secondary);
		if (c == INPUT_ESC) {
			put_tty_queue(LITERAL_ESC, &tty->secondary);
		}
		if (c == '\n') {
			if (LEFT(&tty->secondary) < 2)
				break;
			put_tty_queue(INPUT_ESC, &tty->secondary);
			put_tty_queue(END_REC, &tty->secondary);
		}
	}
}
