/*
 * Input line editing for atty.
 *
 * Copyright (C) 1989 by Kenneth Almquist.  All rights reserved.
 * This file is part of atty, which is distributed under the terms specified
 * by the Atty General Public License.  See the file named LICENSE.
 */

#include <linux/types.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include "tty_ed.h"
/* #include "regex.h" */
#include <linux/ctype.h>
#include <linux/signal.h>
#include <linux/mm.h>
#include <asm/segment.h>


#define HISTFILLER -1		/* indicates blank record in history file */
#define HISTSIZE 4096		/* size of history; MUST be a power of 2 */
#define LOCMASK (HISTSIZE - 1)	/* mask to get offset in history area */

/* flags to histadvance */
#define SKIPEMPTY 01		/* skip empty history lines */
#define SAMEPRGM 02		/* skip history lines read by other programs */
#define HSEARCH 04		/* skip history lines not matching pattern */


struct histhdr {
      short len;		/* length of history line */
      short prevlen;		/* length of preceding line */
      long prgmid;		/* identifies program */
      /* text of line goes here */
};


struct histline {
      short len;		/* length of history line */
      short prevlen;		/* length of preceding line */
      long prgmid;		/* identifies program */
      char *text;		/* text of line */
};


struct tty_struct *tty;		/* structure currently being edited */

extern int version;		/* version of .bindc format */
extern int (*edfunc[])();	/* table of editor functions */


#ifdef __STDC__
int runfmatch(int);
void zapline(void);
void delregion(char *, int);
int dosearch(int);
int extractword(char *, int, int);
int histadvance(int, int);
void addhist(void);
void loadhistline(long);
void gethist(long, struct histline *);
#else
int runfmatch();
void zapline();
void delregion();
int dosearch();
int extractword();
int histadvance();
void addhist();
void loadhistline();
void gethist();
#endif


#define getc(fp)	get_fs_byte(fp++)
#define getsh(fp)	get_fs_word((short *)((fp += 2) - 2))
#define badinit(msg)	{printk(msg); edclose(tty); return;}
#define fread(buf, size, fp)	memcpy_fromfs(buf, fp, size), fp += size
#define malloc(nbytes)	((void *)kmalloc(nbytes, GFP_ATOMIC))
#define free(block)	kfree(block)
#define bcopy(from, to, len)	memcpy(to, from, len)


void
edinit(char *fp, struct tty_struct *tty) {
      int ktsize;
      int stringsize;
      int abbrevsize;

      if (getsh(fp) != BINDMAGIC) {
	    badinit("Bad magic number on key binding file\n");
      }
      if (getc(fp) != version) {
	    badinit(".bindc file needs to be recompiled\n");
      }
      ktsize = getc(fp) << 8;
      stringsize = getsh(fp);
      abbrevsize = getsh(fp);
      if ((tty->keytabs = (unsigned char *)malloc(ktsize)) == NULL)
	    badinit("Out of space\n");
      fread((char *)tty->keytabs, ktsize, fp);
#ifdef notdef
      tty->keyhash = genkeyhash();	/*DEBUG*/
#endif
      fread((char *)tty->syntaxtbl, sizeof tty->syntaxtbl, fp);
      if (stringsize != 0) {
	    if ((tty->edstrings = malloc(stringsize)) == NULL)
		  badinit("Out of space\n");
	    fread(tty->edstrings, stringsize, fp);
      }
      if (abbrevsize != 0) {
	    if ((tty->abbrevtbl = malloc(abbrevsize + 1)) == NULL)
		  badinit("Out of space\n");
	    fread(tty->abbrevtbl, abbrevsize, fp);
	    tty->abbrevtbl[abbrevsize] = '\0';	/* mark end of abbrevs */
      }
      tty->endcurline = tty->curline;
      tty->point = tty->curline;
      tty->curtable = tty->curmode = tty->keytabs;
      gettermcap(tty);
}


void
edclose(struct tty_struct *tty) {
      if (tty->keytabs != NULL) {
	    free(tty->keytabs);
	    tty->keytabs = NULL;
      }
      if (tty->edstrings != NULL) {
	    free(tty->edstrings);
	    tty->edstrings = NULL;
      }
      if (tty->abbrevtbl != NULL) {
	    free(tty->abbrevtbl);
	    tty->abbrevtbl = NULL;
      }
}



void
newttychars() {
      /* nothing to do */
}


void
edflush(struct tty_struct *tty_arg) {
      tty = tty_arg;
      zapline();
      /* also reset state to start of command */
      tty->curtable = tty->keytabs;
      tty->prefixarg = 0;
}


void
inchar(char c, struct tty_struct *tty_arg) {
      int cmd;
      int index;
      int status;
      int pfx;
      int (*func)();

      tty = tty_arg;
      c &= 0177;		/* just in case */
      if (tty->ednextc) {
	    func = tty->ednextc;
	    tty->ednextc = NULL;
	    if ((*func)(c) == 1)
		  tty->needbeep++;
	    goto out;
      }
      if (tty->prefixarg && c >= '0' && c <= '9') {
	    if (tty->prefixarg >= 2)
		  tty->prefixval = 0;
	    tty->prefixarg = 1;
	    tty->prefixval = tty->prefixval * 10 + c - '0';
	    goto out;
      }
      cmd = tty->curtable[c];
      index = tty->curtable[c + 128];
      if (cmd == C_FUNC || cmd == C_INSERT) {
	    if (tty->prefixarg == 0)
		  pfx = 1;
	    else if (tty->prefixarg == 3)
		  pfx = -1;
	    else if (tty->prefixsign)
		  pfx = - tty->prefixval;
	    else
		  pfx = tty->prefixval;
	    tty->cmdcharacter = c;
	    if (cmd == C_FUNC)
		  status = (*edfunc[index])(pfx);
	    else
		  status = insert_string(index, pfx);
	    if (status == 1)
		  tty->needbeep++;
	    if (status != 2)
		  tty->prefixarg = 0;
	    tty->curtable = tty->curmode;
      } else if (cmd == C_PFXTBL) {
	    tty->curtable = tty->keytabs + index * 256;
      } else {
	    tty->needbeep++;
	    tty->curtable = tty->curmode;
      }
out:;
}


void
zapline() {
      tty->point = tty->curline;
      tty->endcurline = tty->curline;
      tty->mark = NULL;
}



universal_argument(int n) {
      if (tty->prefixarg == 0) {
	    tty->prefixval = 1;
	    tty->prefixsign = 0;
      }
      tty->prefixarg = 2;
      tty->prefixval *= 4;
      return 2;
}


digit_argument(int n) {
      tty->prefixarg = 1;
      tty->prefixval = tty->cmdcharacter - '0';
      if (tty->prefixval > 9 || tty->prefixval < 0)
	    tty->prefixval = 0;
      tty->prefixsign = 0;
      return 2;
}


negative_argument(int n) {
      tty->prefixarg = 3;
      tty->prefixval = 0;
      tty->prefixsign = 1;
      return 2;
}


mode_0(int n) {
      tty->curmode = tty->keytabs;
      return 0;
}


mode_1(int n) {
      tty->curmode = tty->keytabs + 256;
      return 0;
}


newline(int n) {
      if (tty->edrecurse)
	    return (*tty->edrecurse)(0);
      if (tty->point == tty->endcurline && expandabbrev() < 0)
	    return 1;
      addhist();
      refresh(tty);
      *tty->endcurline++ = '\n';
      senddata(tty->curline, tty->endcurline - tty->curline, tty);
      movetoend(tty);
      tty->curhist = tty->endhist;
      zapline();
      tty->curmode = tty->keytabs;
      return 0;
}


end_of_file(int n) {
      if (tty->edrecurse)
	    return (*tty->edrecurse)(1);
      if (tty->point == tty->endcurline && expandabbrev() < 0)
	    return 1;
      addhist();
      senddata(tty->curline, tty->endcurline - tty->curline, tty);
      freezedisp(1, tty);
      tty->curhist = tty->endhist;
      zapline();
      tty->curmode = tty->keytabs;
      return 0;
}


eof_or_delete_char(int n) {
      if (tty->endcurline == tty->curline)
	    return end_of_file(n);
      else
	    return delete_char(n);
}


newline_and_insert(int n) {
      if (tty->edrecurse)
	    return 1;
      if (tty->point == tty->endcurline && expandabbrev() < 0)
	    return 1;
      if (tty->curhist == tty->endhist) {
	    tty->lasthistlen = tty->endcurline - tty->curline;
	    bcopy(tty->curline, tty->lasthist, tty->lasthistlen);
      }
      addhist();
      refresh(tty);
      *tty->endcurline++ = '\n';
      senddata(tty->curline, tty->endcurline - tty->curline, tty);
      movetoend(tty);
      zapline();
      tty->curmode = tty->keytabs;
      if (tty->curhist < tty->endhist - HISTSIZE) {
	    tty->curhist = tty->endhist;
	    return 1;
      }
      return histadvance(1, 0);
}


file_complete(int n) {
      return runfmatch('c');
}


list_file_completions(int n) {
      return runfmatch('l');
}


runfmatch(int type) {
      char saveline[MAXLLEN];
      register char *p;
      char *savemark;
      char *savepoint;
      int savelen;
      int status;

      p = tty->point;
      while (p < tty->endcurline && infname(*p))
	    p++;
      if (p == tty->curline || ! infname(p[-1]))
	    return 1;
      savelen = tty->endcurline - tty->curline;
      bcopy(tty->curline, saveline, savelen);
      savemark = tty->mark;
      savepoint = p;
      if (tty->promptset && type == 'l') {
	    refresh(tty);
	    movetoend(tty);
	    tty->promptset = 1;		/* retain tty->prompt */
      }
      tty->point = p;
      delregion(tty->endcurline, 0);
      while (p > tty->curline && infname(p[-1]))
	    p--;
      tty->point = tty->curline;
      delregion(p, 0);
      if (makespace(10)) {
	    status = 1;
	    goto out;
      }
      bcopy("fmatch -x ", tty->curline, 10);
      tty->curline[8] = type;
      if (tty->promptset) {
	    if (makespace(1)) {
		  status = 1;
		  goto out;
	    }
	    tty->curline[0] = '\034';
      } else {
	    refresh(tty);
	    movetoend(tty);
      }
#ifdef notdef /* no longer add to history file */
      atend = (tty->curhist == tty->endhist);
      addhist();
      if (atend)
	    tty->curhist = tty->endhist;
#endif
      *tty->endcurline++ = '\n';
      senddata(tty->curline, tty->endcurline - tty->curline, tty);
      status = 0;
out:
      tty->endcurline = tty->curline + savelen;
      bcopy(saveline, tty->curline, savelen);
      tty->mark = savemark;
      tty->point = savepoint;
      return status;
}

      
tty_intr(int n) {
      if (tty->edrecurse)
	    return (*tty->edrecurse)(1);
      sendsig(SIGINT, tty);
      return 0;
}


tty_quit(int n) {
      sendsig(SIGQUIT, tty);
      return 0;
}


tty_susp(int n) {
      sendsig(SIGTSTP, tty);
      return 0;
}


do_quoted_insert(int c) {
      tty->cmdcharacter = c;
      return self_insert(tty->prefixval);
}


quoted_insert(int n) {
      tty->ednextc = do_quoted_insert;
      tty->prefixval = n;
      return 0;
}


/*
 * Delete the characters between the point and "end".  If "kill" is true,
 * copy to kill buffer.  Exception:  we don't copy to kill if the region
 * is empty.
 */

void
delregion(char *end, int kill) {
      register char *p, *q;
      register char *k;

      if (end > tty->endcurline)
	    end = tty->endcurline;
      if (end < tty->curline)
	    end = tty->curline;
      if (end > tty->point) {
	    p = tty->point;
	    q = end;
      } else {
	    p = end;
	    q = tty->point;
      }
      if (kill && p != q) {
	    end = p;
	    k = tty->killbuffer;
	    while (p < q)
		  *k++ = *p++;
	    tty->killsize = k - tty->killbuffer;
	    p = end;
      }
      tty->point = p;
      if (tty->mark && tty->mark > p) {
	    if (tty->mark > q)
		  tty->mark -= q - p;
	    else
		  tty->mark = p;
      }
      while (q < tty->endcurline)
	    *p++ = *q++;
      tty->endcurline = p;
}


delete_backward_char(int n) {
      delregion(tty->point - n, tty->prefixarg);
      return 0;
}


delete_char(int n) {
      delregion(tty->point + n, tty->prefixarg);
      return 0;
}


kill_word(int n) {
      char *save = tty->point;

      forward_word(n);
      delregion(save, 1);
      return 0;
}


backward_kill_word(int n) {
      char *save = tty->point;

      backward_word(n);
      delregion(save, 1);
      return 0;
}


kill_region(int n) {
      if (tty->mark == NULL)
	    return 1;
      delregion(tty->mark, 1);
      return 0;
}


kill_line(int n) {
      delregion(n <= 0? tty->curline : tty->endcurline, 1);
      return 0;
}


kill_input(int n) {
      tty->point = tty->curline;
      delregion(tty->endcurline, 1);
      return 0;
}


copy_region_as_kill(int n) {
      char *savepoint = tty->point;
      char *savemark = tty->mark;
      delregion(tty->mark, 1);
      yank(1);
      tty->point = savepoint;
      tty->mark = savemark;
      return 0;
}


forward_char(int n) {
      int left;
      left = tty->endcurline - tty->point;
      if (n > left)
	    n = left;
      left = tty->curline - tty->point;
      if (n < left)
	    n = left;
      tty->point += n;
      return 0;
}


backward_char(int n) {
      return forward_char(-n);
}


int
inword(int c) {
      c &= 0177;
      return tty->syntaxtbl[0][c >> 3] & 1 << (c & 07);
}


int
infname(int c) {
      c &= 0177;
      return tty->syntaxtbl[1][c >> 3] & 1 << (c & 07);
}


int
inabbrev(int c) {
      c &= 0177;
      return tty->syntaxtbl[2][c >> 3] & 1 << (c & 07);
}


forward_word(int n) {
      if (n < 0)
	    return backward_word(-n);
      while (--n >= 0) {
	    while (tty->point < tty->endcurline && ! inword(*tty->point))
		  tty->point++;
	    while (tty->point < tty->endcurline && inword(*tty->point))
		  tty->point++;
      }
      return 0;
}


backward_word(int n) {
      if (n < 0)
	    return forward_word(-n);
      while (--n >= 0) {
	    while (--tty->point >= tty->curline && ! inword(*tty->point));
	    tty->point++;
	    while (--tty->point >= tty->curline && inword(*tty->point));
	    tty->point++;
      }
      return 0;
}


end_of_line(int n) {
      tty->point = tty->endcurline;
      return 0;
}


beginning_of_line(int n) {
      tty->point = tty->curline;
      return 0;
}



/*
 * Make space for n characters after the point.  Return 1 on failure.
 */

int
makespace(int n) {
      char *p;

      if (tty->endcurline - tty->curline + n > MAXLLEN)
	    return 1;
      for (p = tty->endcurline ; --p >= tty->point ; )
	    *(p + n) = *p;
      tty->endcurline += n;
      if (tty->mark != NULL && tty->mark > tty->point)
	    tty->mark += n;
      return 0;
}


void
insertchars(char *p, int n, struct tty_struct *tty) {
      if (makespace(n) != 0) {
	    tty->needbeep++;
	    return;
      }
      bcopy(p, tty->point, n);
      tty->point += n;
}


self_insert(int n) {
      if (tty->abbrevtbl && ! inabbrev(tty->cmdcharacter)) {
	    if (expandabbrev() < 0)
		  return 1;
      }
      while (--n >= 0) {
	    if (tty->point == tty->endcurline) {
		  if (tty->endcurline == &tty->curline[MAXLLEN])
			return 1;
		  tty->endcurline++;
	    } else {
		  if (makespace(1) != 0)
			return 1;
	    }
	    *tty->point++ = tty->cmdcharacter;
      }
      return 0;
}


yank(int n) {
      if (makespace(tty->killsize) != 0)
	    return 1;
      bcopy(tty->killbuffer, tty->point, tty->killsize);
      tty->mark = tty->point +  tty->killsize;
      if (! tty->prefixarg)
	    exchange_point_and_mark(0);
      return 0;
}


insert_string(int stringnum, int n) {
      char *p;
      int len;

      p = tty->edstrings;
      while (--stringnum >= 0)
	    p = p + *p + 1;
      if (tty->abbrevtbl && *p != 0 &&! inabbrev(p[1])) {
	    if (expandabbrev() < 0)
		  return 1;
      }
      while (--n >= 0) {
	    if (makespace(len = *p) != 0) {
		  return 1;
	    }
	    bcopy(p + 1, tty->point, len);
	    tty->point += len;
      }
      return 0;
}


set_mark(int n) {
      tty->mark = tty->point;
      return 0;
}


exchange_point_and_mark(int n) {
      char *p;

      if (tty->mark == NULL)
	    return 1;
      p = tty->mark;
      tty->mark = tty->point;
      tty->point = p;
      return 0;
}


/*
 * Convert characters between the point and the specified location to
 * upper case.
 */

void
upcase(char *end) {
      char *p;
      int len;

      if (end > tty->point) {
	    p = tty->point;
	    len = end - tty->point;
      } else {
	    p = end;
	    len = tty->point - end;
      }
      while (--len >= 0) {
	    if (islower(*p))
		  *p = toupper(*p);
	    p++;
      }
}


upcase_char(int n) {
      char *save = tty->point;

      forward_char(n);
      upcase(save);
      return 0;
}


upcase_word(int n) {
      char *save = tty->point;

      forward_word(n);
      upcase(save);
      return 0;
}


upcase_region(int n) {
      upcase(tty->mark);
      return 0;
}


gosling_transpose_chars(int n) {
      char c;

      if (tty->prefixarg)
	    return transpose_chars(n);
      if (tty->point < tty->curline + 2)
	    return 1;
      c = tty->point[-1];
      tty->point[-1] = tty->point[-2];
      tty->point[-2] = c;
      return 0;
}


transpose_chars(int n) {
      char c;

      if (tty->point == tty->endcurline && tty->point >= tty->curline + 2 && ! tty->prefixarg)
	    tty->point--;
      if (tty->point <= tty->curline || n > 0 && tty->point >= tty->endcurline)
	    return 1;
      c = *(tty->point - 1);
      delregion(tty->point - 1, 0);
      forward_char(n);
      makespace(1);
      *tty->point++ = c;
      return 0;
}


transpose_words(int n) {
      char *save = tty->point;
      char *start;
      char *loc;
      int end1, start2, end2;
      char buf[MAXLLEN];

      if (tty->point == tty->endcurline)
	    backward_word(1);
      loc = tty->point;
      backward_word(1);
      start = tty->point;
      if (start >= loc || ! inword(*start))
	    goto fail;
      forward_word(1);
      if (tty->point > loc)
	    tty->point = loc;
      end1 = tty->point - start;
      forward_word(1);
      end2 = tty->point - start;
      backward_word(1);
      if (tty->point < loc) {
	    tty->point = loc;
	    if (! inword(*loc))
		  goto fail;
      }
      start2 = tty->point - start;
      bcopy(start, buf, end2);
      tty->point = start + start2;
      delregion(start + end2, 0);
      tty->point = start;
      delregion(start + end1, 0);
      makespace(end2 - start2);
      bcopy(buf + start2, start, end2 - start2);
      tty->point = start + end2 - end1;
      makespace(end1);
      bcopy(buf, tty->point, end1);
      tty->point += end1;
      if (save == tty->endcurline)
	    tty->point = save;
      return 0;

fail:
      tty->point = save;
      return 1;
}


re_search_backward(int n) {
      return re_search_forward(-n);
}


re_search_forward(int n) {
      if (tty->edrecurse)
	    return 1;
      tty->edrecurse = dosearch;
      tty->savecurlinelen = tty->endcurline - tty->curline;
      bcopy(tty->curline, tty->savecurline, tty->savecurlinelen);
      tty->savepoint = tty->point;
      tty->savemark = tty->mark;
      tty->saveprefixval = n;
      zapline();
      tty->editprompt = tty->cmdcharacter? tty->cmdcharacter : '/';
      tty->curmode = tty->keytabs;
      return 0;
}


dosearch(int flag) {
      int len;
      int i;

      tty->edrecurse = NULL;
      tty->editprompt = '\0';
      if (flag)
	    goto fail;
      len = tty->endcurline - tty->curline;
      if (len == 0) {
	    if ((len = tty->savepatlen) == 0)
		  goto fail;
	    bcopy(tty->savepat, tty->curline, len);
      }
      tty->curline[len] = '\0';
      if (re_compile(tty->curline, len, tty) == 0) {
	    goto fail;
      }
      if (len >= sizeof tty->savepat) {
	    tty->savepatlen = 0;
      } else {
	    tty->savepatlen = len;
	    bcopy(tty->curline, tty->savepat, len);
      }
      bcopy(tty->savecurline, tty->curline, tty->savecurlinelen);
      tty->endcurline = tty->curline + tty->savecurlinelen;
      tty->point = tty->savepoint;
      tty->mark = tty->savemark;
      i = histadvance(tty->saveprefixval, HSEARCH);
      freepat();
      return i;
fail:
      bcopy(tty->savecurline, tty->curline, tty->savecurlinelen);
      tty->endcurline = tty->curline + tty->savecurlinelen;
      tty->point = tty->savepoint;
      tty->mark = tty->savemark;
      return 1;
}


get_history_word(int n) {
      struct histline hline;

      if (tty->endhist == 0)
	    return 1;
      gethist(tty->endhist - tty->histprevlen - sizeof (struct histhdr), &hline);
      if (! tty->prefixarg)
	    n = -1;
      return extractword(hline.text, hline.len, n);
}


last_output_line(int n) {
      if (! tty->prefixarg)
	    n = 0;
      return extractword(tty->lastoutline, tty->lastoutlinelen, n);
}


/*
 * Insert the contents of word n of the line.  If n is zero, insert the
 * entire line; if it is negative, count words from the end.  In this
 * routine, words are always delimited by spaces and tabs.
 */

int
extractword(line, len, n)
      char *line;
      {
      char *p;
      char *q;
      char *endline = line + len;
      int wordlen;

      if (n > 0) {
	    p = line;
	    for (;;) {
		  while (p < endline && ! infname(*p))
			p++;
		  q = p;
		  while (q < endline && infname(*q))
			q++;
		  if (--n == 0)
			break;
		  p = q;
	    }
      } else if (n < 0) {
	    q = endline;
	    for (;;) {
		  while (q > line && ! infname(q[-1]))
			q--;
		  p = q;
		  while (p > line && infname(p[-1]))
			p--;
		  if (++n == 0)
			break;
		  q = p;
	    }
      } else { /* n == 0 */
	    p = line;
	    q = endline;
      }
      wordlen = q - p;
      if (makespace(wordlen) != 0)
	    return 1;
      bcopy(p, tty->point, wordlen);
      tty->mark = tty->point;
      tty->point += wordlen;
      return 0;
}


next_history(int n) {
      return histadvance(n, SKIPEMPTY);
}


previous_history(int n) {
      return histadvance(-n, SKIPEMPTY);
}


end_of_history(int n) {
      loadhistline(tty->endhist);
      return 0;
}


beginning_of_history(int n) {
      histadvance(-32767, 0);
      return 0;
}


/*
 * Move through the history file, and load the result into tty->curline.
 * N is the number of items to move:  a negative value of n moves
 * backward in the history file and a postive value of n moves forward.
 * The flags control which history lines are counted.  The search starts
 * from the location specified by the tty->curhist variable.
 *
 * The structure of the history file is as follows.  Each line in the
 * history file consists of a header, defined as (struct histhdr), followed
 * by the text of the line.  Locations in the history file are identified
 * by the number of bytes written since the file was first created.  To
 * keep the file from growing without bound, the file wraps around at
 * HISTSIZE bytes.  Thus any location less than (endhist - HISTSIZE) has
 * been overwritten.  (Endhist points to the end of the history file.)
 */

int
histadvance(n, flags) {
      struct histline hline;
      long loc;
      long found;
      int back;
      int oldloc;
      int prgmid = 1;

      if (tty->edrecurse)
	    return 1;
      loc = tty->curhist;
      if (loc == tty->endhist) {
	    tty->lasthistlen = tty->endcurline - tty->curline;
	    bcopy(tty->curline, tty->lasthist, tty->lasthistlen);
      }
      back = 0;
      if (n < 0) {
	    n = -n;
	    back = 1;
      }
      found = loc;
      while (n >= 0) {
	    oldloc = loc;
	    if (back) {
		  if (loc == tty->endhist) {
			loc -= sizeof (struct histhdr) + tty->histprevlen;
		  } else if (loc < 0 || loc < tty->endhist - HISTSIZE) {
			break;
		  } else {
			gethist(loc, &hline);
			loc -= sizeof (struct histhdr) + hline.prevlen;
		  }
	    } else {
		  if (loc == tty->endhist) {
			loc++;
		  } else if (loc > tty->endhist) {
			break;
		  } else {
			gethist(loc, &hline);
			loc += sizeof (struct histhdr) + hline.len;
		  }
	    }
	    if (oldloc == tty->endhist) {
		  hline.prgmid = prgmid;
		  hline.text = tty->lasthist;
		  hline.len = tty->lasthistlen;
	    }
	    if (hline.prgmid != HISTFILLER
	     && ((flags & SKIPEMPTY) == 0 || oldloc == tty->endhist || hline.len > 0)
	     && ((flags & SAMEPRGM) == 0 || hline.prgmid == prgmid)
	     && ((flags & HSEARCH) == 0 || oldloc == tty->curhist ||
			re_match(hline.text, hline.len, tty))) {
		  n--;
		  found = oldloc;
	    }
      }
      if (found != tty->curhist || n < 0)
	    loadhistline(found);
      return n >= 0? 1 : 0;
}


/*
 * Add the current line to the history file.  If echoing is turned off,
 * we store a blank line rather than the actual line so that we don't
 * have passwords in the history file.
 */

void
addhist() {
      struct histhdr hdr;
      int space;
      int len = L_ECHO(tty)? tty->endcurline - tty->curline : 0;
      char *lp;

      hdr.prevlen = tty->histprevlen;
      space = HISTSIZE - (tty->endhist & LOCMASK);
      if (space < len + 2 * sizeof (struct histhdr)) {
	    hdr.len = space - sizeof (struct histhdr);
	    hdr.prgmid = HISTFILLER;
	    bcopy((char *)&hdr, tty->histarea + (tty->endhist & LOCMASK), sizeof hdr);
	    hdr.prevlen = hdr.len;
	    tty->endhist += space;
      }
      lp = tty->histarea + (tty->endhist & LOCMASK);
      hdr.len = len;
      hdr.prgmid = 0;		/* for now */
      bcopy((char *)&hdr, lp, sizeof hdr);
      bcopy(tty->curline, lp + sizeof hdr, len);
      tty->histprevlen = len;
      tty->endhist += sizeof hdr + len;
}


/*
 * Load the contents of a history line into tty->curline.
 */

void
loadhistline(loc)
      long loc;
      {
      struct histline hline;

      if (loc == tty->endhist) {
	    if (tty->curhist == tty->endhist) {
		  return;
	    }
	    hline.text = tty->lasthist;
	    hline.len = tty->lasthistlen;
      } else {
	    gethist(loc, &hline);
      }
      bcopy(hline.text, tty->curline, hline.len);
      tty->endcurline = tty->curline + hline.len;
      tty->point = tty->endcurline;
      tty->mark = NULL;
      tty->curhist = loc;
}


/*
 * Read a history entry into hline.  To make the history file be a real
 * file rather than a block of memory, change this routine and the addhist
 * routine.
 */

void
gethist(long loc, struct histline *hline) {
      char *lp = tty->histarea + (loc & LOCMASK);

      bcopy(lp, (char *)hline, sizeof (struct histhdr));
      hline->text = lp + sizeof (struct histhdr);
}


/*
 * Expand the abbreviation before point, if any.  Returns 1 if expansion
 * occurred, 0 if no expansion occurred, and -1 if the expansion could
 * not be performed because there was not sufficient space.
 */

int
expandabbrev() {
      char *start;
      int len;
      char *p;

      if (tty->abbrevtbl == NULL)
	    return 0;
      start = tty->point;
      while (--start >= tty->curline && inabbrev(*start));
      start++;
      if (start == tty->point)
	    return 0;
      len = tty->point - start;
      p = tty->abbrevtbl;
      for (;;) {
	    if (*p == 0)
		  return 0;		/* end of table */
	    if (*p == len && memcmp(p + 1, start, len) == 0)
		  break;
	    p += *p + 1;
	    p += *p + 1;
      }
      p += *p + 1;
      if (tty->endcurline - tty->curline + *p - len > MAXLLEN)
	    return -1;
      delregion(start, 0);
      if (makespace(*p))
	    return -1;			/* can't happen */
      bcopy(p + 1, tty->point, *p);
      tty->point += *p;
      return 1;
}


/* DEBUG */

int
genkeyhash() {
      int hash;
      unsigned char *p;

      hash = 123456; p = tty->keytabs + 256;
      do { hash = ((hash << 8) + *--p) % 8380087; } while (p >= tty->keytabs);
      return hash;
}


void
chkkeytab() {
      int hash;
      unsigned char *p;

      hash = 123456; p = tty->keytabs + 256;
      do { hash = ((hash << 8) + *--p) % 8380087; } while (p >= tty->keytabs);
      if (hash != tty->keyhash)
	    printk("Bad key table\n");
}

/* END DEBUG */
