/*
 * Copyright (C) 1989 by Kenneth Almquist.
 */

#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include "tty_ed.h"

#define bcopy(from, to, len)	memcpy(to, from, len)


/*
 * All output passes through the outchars routine.  The routine keeps
 * track of the current column and identifies escape sequences that
 * are interpreted by atty itself.
 */
void
tty_put_char(char c, struct tty_struct *tty) {
      int newcol;
      int i;
      char *q;

      if (O_POST(tty)) {
	    if (c == '\r') {
		  if (O_CRNL(tty))
			c = '\n';
		  else if (O_NOCR(tty) && tty->column == 0)
			return;
	    } else if (c == '\n' && O_NLCR(tty)) {
		  int saveflag = tty->termios->c_oflag;
		  tty->termios->c_oflag &=~ OCRNL;
		  tty_put_char('\r', tty);
		  tty->termios->c_oflag = saveflag;
	    } else if (c == '\t' && (tty->termios->c_oflag & TABDLY) == TAB3) {
		  i = 8 - (tty->column & 07);
		  do tty_put_char(' ', tty);
		  while (--i != 0);
		  return;
	    }
#ifdef notdef
	    if (O_LCUC(tty))
		  c = toupper(c);
#endif
      }
      switch (tty->outstate) {
      case 0:
      state0:	/* base state */
	    if (c >= ' ' && c <= '~') {
		  if (tty->column < COLUMNS)
			tty->outline[tty->column] = c;
		  tty->column++;
	    } else if (c == '\t') {
		  newcol = (tty->column + 8) &~ 07;
		  while (tty->column < newcol && tty->column < COLUMNS)
			tty->outline[tty->column++] = ' ';
		  tty->column = newcol;
	    /* may need to fix the handling of return */
	    } else if (c == '\r' || (c == '\n' && O_NLRET(tty))) {
		  if ((tty->lastoutlinelen = tty->column) > COLUMNS)
			tty->lastoutlinelen = COLUMNS;
		  bcopy(tty->outline, tty->lastoutline, tty->lastoutlinelen);
		  tty->column = 0;
	    } else if (c == '\033') {	/* ESCAPE */
		  tty->outstate = 1;
	    } else if (c == '\b') {
		  if (tty->column > 0)
			tty->column--;
	    }
	    break;
      case 1:
      state1:	/* escape character seen */
	    if (c == '[') {
		  tty->outstate = 2;
		  tty->nparam = 0;
		  tty->param[0] = tty->param[1] = 0;
		  tty->curnumber = NULL;
	    } else if (c == ']' && tty->keytabs) {
		  tty->outstate = 3;
	    } else if (c == 'H') {	/* ESC H = home on some terminals */
		  tty->column = 0;
		  tty->outstate = 0;
	    } else {
		  tty->outstate = 0;
	    }
	    break;
      case 2:
      state2:	/* ESC [ seen */
	    if (c >= '0' && c <= '9') {
		  if (tty->curnumber != NULL) {
			*tty->curnumber = *tty->curnumber * 10 + c - '0';
		  } else if (tty->nparam < 2) {
			tty->curnumber = &tty->param[tty->nparam++];
			*tty->curnumber = c - '0';
		  }
	    } else if (c == '?') {
		  /* ignore it for now */
	    } else if (c == ';') {
		  tty->curnumber = NULL;
	    } else {
		  goto gotescape;
	    }
	    break;

gotescape:
	    /* got a complete escape sequence */
	    switch (c) {
		  case 'H':
			tty->column = tty->param[1] > 0? tty->param[1] - 1 : 0;
			break;
		  case 'C':
			tty->column += tty->param[0];
			break;
		  case 'D':
			if ((tty->column -= tty->param[0]) < 0)
			      tty->column = 0;
			break;
		  case 'L':  case'M':
			tty->column = 0;
			break;
		  case 'J':
			if (tty->param[0] == 1 || tty->param[0] == 2) {
			      for (i = 0 ; i < tty->column && i < COLUMNS; i++)
				    tty->outline[i] = ' ';
			}
			break;
	    }
	    tty->outstate = 0;
	    break;
      case 3:
      state3:	/* got ESC ], which introduces an atty control sequence */
	    if (c == 'P' || c == 'I' || c == 'D') {
		  tty->esccmd = c;
		  tty->strparamp = tty->strparam;
		  tty->outstate = 4;
	    } else {
		  tty->outstate = 0;
	    }
	    break;
      case 4:
      state4:	/* reading tty->prompt following ESC ] P */
	    if (c == '\n') {
		  *tty->strparamp = '\0';
		  goto myesc;
	    }
	    if (c != '\r' && tty->strparamp < &tty->strparam[255]) {
		  *tty->strparamp++ = c;
	    }
	    break;

myesc:
	    /* got an ESC ] ... sequence */
	    switch (tty->esccmd) {
	    case 'P':
		  if (tty->promptset) {
			freezedisp(0, tty);
#ifdef notdef
			if (tty->column != 0) {
			      ttyoutc('\r', tty);
			      ttyoutc('\n', tty);
			}
			for (tty->column = 0 ; tty->prompt[tty->column] ; tty->column++)
			      ttyoutc(tty->prompt[tty->column], tty);
			bcopy(tty->prompt, tty->outline, tty->column);
#endif
		  }
		  q = tty->strparam;
		  i = 0;
		  while (*q >= '0' && *q <= '9') {
			i = 10 * i + *q - '0';
			q++;
		  }
		  if (*q == ';')
			q++;
		  tty->promptcode = i;
		  if (strlen(q) > COLUMNS - 1)
      			q[COLUMNS - 1] = '\0';
		  strcpy(tty->prompt, q);
		  tty->promptset = 1;
		  break;
	    case 'D':
		  tty->promptset = 0;
		  break;
	    case 'I':
		  insertchars(tty->strparam, strlen(tty->strparam), tty);
		  break;
	    }
	    tty->outstate = 0;
	    return;		/* no output */
      }
      if (tty->outstate <= 3) {
#ifdef notdef
	    ttyoutc(c, tty);
#else
	    put_tty_queue(c, &tty->write_q);
#endif
      }
}
