/*
 * Version of regular expression matching commands by Kenneth Almquist.
 */

#include <linux/types.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include "tty_ed.h"
#include "softspace.h"


#define TRUE	1
#define FALSE	0

#undef BCHAR	/* if defined do fast check of first character */

/* Compiled version of regular expressions */
#define PCHAR	0		/* match a single character */
#define PLET	1		/* match a letter, lower or upper case */
#define PDOT	2		/* match any character except newline */
#define PCCL	3		/* character class */
#define PSTAR	4		/* zero or more occurances of next pattern */
#define PBOL	5		/* match beginning of line */
#define PEOL	6		/* match end of line */
#define PBEGWORD 7		/* match beginning of word */
#define PENDWORD 8		/* match end of word */
#define PENDPAT	9		/* end of pattern */

const char	patextra[] = {1, 1, 0, 32, 0, 0, 0, 0, 0};


#define ARB 4

struct pat {
	struct pat     *next;
	char            pattext[ARB];
};

#ifdef BCHAR
char           *bchar = NULL;
#endif
struct pat     *patlist = NULL;

void            freepat();

#define CHARMASK(c)	((unsigned char)(c))
#define ISWORD(c)	inword(c)
#define ewprintf(msg)
#define ckalloc(size)	kmalloc(size, GFP_ATOMIC)
#define free(block)	kfree(block)
#define bcopy(from, to, nbytes)	memcpy(to, from, nbytes);
#define assert(condition)	if (! (condition))  panic("assert failed: " #condition)



/*
 * Parse a regular expression.  Returns TRUE on success, FALSE on failure.
 * The first argument is a pattern string.  The second argument is the
 * length of the string.  The pattern string must be terminated with a nul
 * character, but may also contain nul characters.  Patlen is used to
 * determine whether a nul character is part of the string or is the
 * string terminator.
 */
int
re_compile(pattern, patlen, tty)
	char           *pattern;
	struct tty_struct *tty;
{
	char           *outp;
	char           *lastout;
	char           *startpat;
	char           *endpat;
	int             i;
	int             not;
	int             c, c2;
	struct pat     *pp;
#ifdef BCHAR
	int             firstchar;
#endif

	freepat();			/* just in case not called earlier */
#ifdef BCHAR
	if ((bchar = ckalloc(256)) == NULL) {
		return FALSE;
	}
	memset(bchar, 0, 256);
#endif
	endpat = pattern + patlen;
	patlist = NULL;
	do {  /* until end of pattern */
		outp = softint_space;
		startpat = pattern;
#ifdef BCHAR
		firstchar = TRUE;
#endif
		for (;;) {
			if (outp >= softint_space + sizeof softint_space - 4) {
				ewprintf("Pattern too long");
				goto bad;
			}
			lastout = outp;
			switch (*pattern++) {
			case '^':
				if (pattern != startpat + 1)
					goto dft;
				*outp++ = PBOL;
				break;
			case '$':
				if (pattern != endpat && (pattern[0] != '\\' || pattern[1] != '|'))
					goto dft;
				*outp++ = PEOL;
				break;
			case '\\':
				if (pattern == endpat)
					goto incomplete;
				switch (*pattern++) {
				case '<':
					*outp++ = PBEGWORD;
					break;
				case '>':
					*outp++ = PENDWORD;
					break;
				case '|':
					goto breakfor;
				default:
					goto dft;
				}
				break;
			case '.':
				*outp++ = PDOT;
#ifdef BCHAR
				if (firstchar) {
					i = bchar['\n'];
					memset(bchar, 1, 256);
					bchar['\n'] = i;
				}
#endif
checkstar:
				if (*pattern == '*') {
					pattern++;
					i = outp - lastout;
					do {
						*outp = *(outp - 1);
						outp--;
					} while (outp != lastout);
					*outp = PSTAR;
					outp += 1 + i;
#ifdef BCHAR
				} else {
					firstchar = FALSE;
#endif
				}
				break;
			case '[':
				*outp++ = PCCL;
				i = 32;
				do
					*outp++ = 0;
				while (--i > 0);
				outp -= 32;
				not = 0;
				if (*pattern == '^') {
					not = 1;
					pattern++;
				}
				do {
					if (pattern >= endpat) {
						ewprintf("Unmatched [");
						goto bad;
					}
					c = c2 = CHARMASK(*pattern++);
					if (*pattern == '-' && pattern[1] != ']') {
						c2 = CHARMASK(pattern[1]);
						pattern += 2;
					}
					if (c2 < c) {
						ewprintf("Bad character range");
						goto bad;
					}
					while (c <= c2) {
						outp[c >> 3] |= 1 << (c & 07);
						c++;
					}
				} while (*pattern != ']');
				pattern++;	/* skip ']' */
				assert(outp[-1] == PCCL);
				if (not) {
					i = 32;
					do {
						*outp = ~*outp;
						outp++;
					} while (--i > 0);
				} else {
					outp += 32;
				}
#ifdef BCHAR
				if (firstchar) {
					for (i = 0; i < 256; i++) {
						if (outp[(i >> 3) - 32] & 1 << (i & 07))
							bchar[i] = TRUE;
					}
				}
#endif
				goto checkstar;
			case '\0':
				if (pattern > endpat)
					goto breakfor;
				goto dft;
			dft: default:
				c = pattern[-1];
#ifdef CASEFOLDSEARCH
				if (CASEFOLDSEARCH) {
					if (ISLOWER(c)) {
						*outp++ = PLET;
					} else if (ISUPPER(c)) {
						c = TOLOWER(c);
						*outp++ = PLET;
					} else {
						*outp++ = PCHAR;
					}
				} else
#endif
					*outp++ = PCHAR;
				*outp++ = c;
#ifdef BCHAR
				if (firstchar) {
					bchar[CHARMASK(c)] = TRUE;
#ifdef CASEFOLDSEARCH
					if (CASEFOLDSEARCH && ISLOWER(c))
						bchar[TOUPPER(c)] = TRUE;
#endif
				}
#endif
				goto checkstar;
			}
		}
breakfor:	;
		*outp++ = PENDPAT;
#ifdef BCHAR
		if (firstchar)
			memset(bchar, 1, 256);
#endif
		if ((pp = (struct pat *) ckalloc(sizeof *pp - ARB + (outp - softint_space))) == NULL) {
			goto bad;
		}
		pp->next = patlist;
		patlist = pp;
		bcopy(softint_space, (char *) pp->pattext, outp - softint_space);
	} while (pattern <= endpat);
	return TRUE;

incomplete:
	ewprintf("Incomplete regular expression");
bad:
	freepat();
	return FALSE;
}


/*
 * Free the parsed pattern created by a call to re_compile.
 */
void
freepat()
{
	register struct pat *pp, *prev;

#ifdef BCHAR
	if (bchar != NULL) {
		free(bchar);
		bchar = NULL;
	}
#endif
	for (pp = patlist; pp;) {
		prev = pp;
		pp = pp->next;
		free((char *) prev);
	}
	patlist = NULL;
}
	

/*
 */
int
re_match(text, len, tty)
	char	       *text;
	int		len;
	struct tty_struct *tty;
{
	register int    c;
	register int	offset;

	offset = 0;
	do {
#ifdef BCHAR
		c = len == 0? '\n' : *text;
		if (bchar[c] && matchpat(text, len, offset)) {
#else
		if (matchpat(text, len, offset)) {
#endif
			return TRUE;
		}
	} while (++offset <= len);
	return FALSE;
}


int
matchpat(text, len, offset)
	char	       *text;
	int		len;
{
	register struct pat *pp;
	int             redo;

	for (pp = patlist; pp; pp = pp->next) {
		if (matchalt(pp->pattext, text, len, offset)) {
			return TRUE;
		}
	}
	return FALSE;
}


int
matchalt(pat, text, len, offset)
	char           *pat;
	char	       *text;
{
	int             star;
	int             begstar;
	int             c;

	star = 0;
	for (;;) {
		if (offset >= len)
			c = -1;
		else
			c = text[offset];
		switch (*pat++) {
		case PCHAR:
			/* trace("PCHAR %C", CHARMASK(*pat)); */
			if (c != CHARMASK(*pat))
				goto nomatch;
			break;
#ifdef CASEFOLDSEARCH
		case PLET:
			/* trace("PLET %C", CHARMASK(*pat)); */
			if (c != *pat && c != TOUPPER(*pat))
				goto nomatch;
			break;
#endif
		case PDOT:
			/* trace("PDOT"); */
			if (c < 0 || c == '\n')
				goto nomatch;
			break;
		case PCCL:
			/* trace("PCCL"); */
			if (c < 0 || (pat[(c & 0377) >> 3] & 1 << (c & 07)) == 0)
				goto nomatch;
			break;
		case PBOL:
			/* trace("PBOL"); */
			if (offset != 0)
				goto nomatch;
			continue;		/* matches no characters */
		case PEOL:
			/* trace("PEOL"); */
			if (offset != len)
				goto nomatch;
			continue;		/* matches no characters */
		case PBEGWORD:
			if (offset > 0 && ISWORD(text[offset - 1]))
				goto nomatch;
			if (c < 0 || !ISWORD(c))
				goto nomatch;
			continue;		/* matches no characters */
		case PENDWORD:
			if (offset == 0 || !ISWORD(text[offset - 1]))
				goto nomatch;
			if (c >= 0 && ISWORD(c))
				goto nomatch;
			continue;		/* matches no characters */
		case PSTAR:
			/* trace("PSTAR"); */
			star = 1;
			begstar = offset;
			continue;		/* matches no characters */
		case PENDPAT:
			/* trace("PENDPAT offset=%d", offset); */
			return TRUE;
		default:
			panic("Pattern error");
		}
		if (star) {
			pat--;
		} else {
			pat += patextra[*(pat - 1)];
		}
		offset++;
	}
nomatch:
	if (!star)
		return FALSE;
	pat += patextra[*(pat - 1)];
	while (offset >= begstar) {
		if (matchalt(pat, text, len, offset))
			return TRUE;
		offset--;
	}
	return FALSE;
}
