/*
 *  linux/kernel/chr_drv/tty/tty.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */


/*
 * This file contains routines shared by all tty devices.
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/timer.h>
#include <linux/ctype.h>
#include <linux/kd.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/keyboard.h>
#include <linux/interrupt.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/bitops.h>

#define QUEUE_BLOCK_SIZE (TTY_BUF_SIZE * sizeof (unsigned short))

struct tty_struct *tty_table[MAX_TTYS];
struct termios *tty_termios[MAX_TTYS];	/* We need to keep the termios state */
					/* around, even when a tty is closed */
struct tty_struct *tty_list;		/* list of open ttys */
int tty_check_write[MAX_TTYS/32];	/* bitfield for the bh handler */
int tty_debug = 0;			/* if set, turn on tracing */
#ifdef notdef
#define kfree_s(block, size)	printk("<kfree_x(0x%X, %d)>", block, size)
#define free_page(page)		printk("<free_page(0x%X)>", page)
#endif

static void initialize_tty_struct(int line, struct tty_struct *tty);
static void initialize_termios(int line, struct termios *tp);

static void tty_timer(void *);
static void tty_trigger_timer(void);

static int tty_read(struct inode *, struct file *, char *, int);
static int tty_write(struct inode *, struct file *, char *, int);
static int tty_select(struct inode *, struct file *, int, select_table *);
static int tty_open(struct inode *, struct file *);
static void tty_release(struct inode *, struct file *);

/*
 * The line discipline table.  The routines tty_add_ldisc and tty_delete_disc
 * are used to create and remove entries from this table.
 */
extern struct ldisc tty_ldisc_0;
extern struct ldisc crted_ldisc;

struct ldisc *ldiscs[NR_LDISCS] = {
	&tty_ldisc_0,
	&crted_ldisc,
};


int tty_add_ldisc(struct ldisc *ldisc)
{
	register struct ldisc **ldpp;

	for (ldpp = ldiscs ; ; ldpp++) {
		if (ldpp >= &ldiscs[NR_LDISCS])
			return -ENOSPC;
		if (*ldpp == NULL)
			break;
		if ((*ldpp)->name == ldisc->name)
			return -EBUSY;
	}
	*ldpp = ldisc;
	return 0;
}

int tty_delete_ldisc(struct ldisc *ldisc)
{
	register struct ldisc **ldpp, **ldpp2;

	for (ldpp = ldpp2 = ldiscs ; ldpp < &ldiscs[NR_LDISCS] ; ldpp++) {
		if (*ldpp != ldisc) {
			*ldpp2++ = *ldpp;
		}
	}
	if (ldpp2 == ldpp)
		return -EINVAL;
	for ( ; ldpp2 < ldpp ; ldpp2++)
		*ldpp2 = NULL;
	return 0;
}


void put_tty_queue(char c, struct tty_queue * queue)
{
	int head;
	unsigned long flags;

	__asm__ __volatile__("pushfl ; popl %0 ; cli":"=r" (flags));
	head = (queue->head + 1) & (TTY_BUF_SIZE-1);
	if (head != queue->tail) {
		queue->buf[queue->head] = c;
		queue->head = head;
	}
	__asm__ __volatile__("pushl %0 ; popfl"::"r" (flags));
}

int get_tty_queue(struct tty_queue * queue)
{
	int result = -1;
	unsigned long flags;

	__asm__ __volatile__("pushfl ; popl %0 ; cli":"=r" (flags));
	if (queue->tail != queue->head) {
		result = queue->buf[queue->tail];
		queue->tail = (queue->tail + 1) & (TTY_BUF_SIZE-1);
	}
	__asm__ __volatile__("pushl %0 ; popfl"::"r" (flags));
	return result;
}


/*
 * This routine copies out a maximum of buflen characters from the
 * read_q; it is a convenience for line disciplins so they can grab a
 * large block of data without calling get_tty_char directly.  It
 * returns the number of characters actually read.
 */
int tty_read_raw_data(struct tty_struct *tty, unsigned char *bufp, int buflen)
{
	int	result = 0;
	unsigned char	*p = bufp;
	unsigned long flags;
	int head, tail;
	
	__asm__ __volatile__("pushfl ; popl %0 ; cli":"=r" (flags));
	tail = tty->read_q.tail;
	head = tty->read_q.head;
	while ((result < buflen) && (tail!=head)) {
		*p++ =  tty->read_q.buf[tail++];
		tail &= TTY_BUF_SIZE-1;
		result++;
	}
	tty->read_q.tail = tail;
	__asm__ __volatile__("pushl %0 ; popfl"::"r" (flags));
	return result;
}


/*
 * Start output on the tty.  This should be called after data has
 * been placed in tty->write_q.
 */
void tty_write_flush(struct tty_struct * tty)
{
	if (!tty) {		/*DEBUG*/
		printk("tty_write called with NULL tty\n");
		return;
	}
	tty->la.write_flush(tty);
}


/*
 * Called to process data in the read_q and copy it (after processing)
 * into the secondary_q.  The code to disable software interrupts is
 * a temporary measure for the pty code.
 */
void tty_read_flush(struct tty_struct * tty)
{
	if (!tty)
		return;
	SOFT_INT_OFF;
	tty->lb.process_input(tty);
	SOFT_INT_ON;
}


static void tty_trigger_timer()
{
	add_timer((HZ + 5) / 10, tty_trigger_timer);
	mark_bh(TTYTIMER_BH);
}

static void tty_timer(void *garbage)
{
	struct tty_struct *tty;
	int i;		/*TEMP*/

	for (tty = tty_list ; tty ; tty = tty->next) {
		for (i = 0 ; tty_table[i] != tty ; i++) {	/*TEMP*/
			if (i >= MAX_TTYS - 1) {
				printk("tty 0x%X not in table\n", tty);
				return;
			}
		}
		if (tty->lb.timer)
			(*tty->lb.timer)(tty);
	}
}

/*
 * Once a tty has been hung up, further I/O operations are prohibited,
 * even if the tty is connected again.  We implement this by changing
 * the fops structure for all people who have the tty open when it is
 * hung up.
 */
static int hung_up_tty_read(struct inode * inode, struct file * file, char * buf, int count)
{
	return 0;
}

static int hung_up_tty_write(struct inode * inode, struct file * file, char * buf, int count)
{
	return -EIO;
}

static int hung_up_tty_select(struct inode * inode, struct file * filp, int sel_type, select_table * wait)
{
	return 1;
}

/*
 * Seeks on a tty always fail.
 */
static int tty_lseek(struct inode * inode, struct file * file, off_t offset, int orig)
{
	return -ESPIPE;
}

static struct file_operations tty_fops = {
	tty_lseek,
	tty_read,
	tty_write,
	NULL,		/* tty_readdir */
	tty_select,
	tty_ioctl,
	NULL,		/* tty_mmap */
	tty_open,
	tty_release
};

static struct file_operations hung_up_tty_fops = {
	tty_lseek,
	hung_up_tty_read,
	hung_up_tty_write,
	NULL,		/* hung_up_tty_readdir */
	hung_up_tty_select,
	tty_ioctl,
	NULL,		/* hung_up_tty_mmap */
	tty_open,
	tty_release
};


/*
 * This routine should be called when carrier detect drops.
 */

void tty_hangup(struct tty_struct * tty)
{
	if (tty == NULL) {
		printk("tty_hangup: called with NULL argument\n");
		return;
	}
	tty->lb.hangup(tty);
}

/*
 * The default value for lb.hangup.  We change the file operators
 * to prevent further reads or writes on the tty.
 */
void tty_std_hangup(struct tty_struct * tty)
{
	struct file * filp;
	struct task_struct **p;
	int dev;

	TTY_TRACE(("tty_hangup called\n"));
	if (!tty)
		return;
	dev = 0x0400 + tty->line;
	filp = file_table + NR_FILE;
	while (filp-- > file_table) {
		if (!filp->f_count)
			continue;
		if (filp->f_rdev != dev)
			continue;
		if (filp->f_op != &tty_fops)
			continue;
		filp->f_op = &hung_up_tty_fops;
	}
	wake_up_interruptible(&tty->secondary.proc_list);
	wake_up_interruptible(&tty->read_q.proc_list);
	wake_up_interruptible(&tty->write_q.proc_list);
	if (tty->session > 0)
		kill_sl(tty->session,SIGHUP,1);
	tty->session = 0;
	tty->pgrp = -1;
 	for (p = &LAST_TASK ; p > &FIRST_TASK ; --p) {
		if ((*p) && (*p)->tty == tty->line)
			(*p)->tty = -1;
	}
	/* Should flush input and output queues -- KA */
}

void tty_unhangup(struct file *filp)
{
	filp->f_op = &tty_fops;
}

int hung_up(struct file * filp)
{
	return filp->f_op == &hung_up_tty_fops;
}

int is_ignored(int sig)
{
	return ((current->blocked & (1<<(sig-1))) ||
	        (current->sigaction[sig-1].sa_handler == SIG_IGN));
}


static int tty_read(struct inode * inode, struct file * file, char * buf, int count)
{
	int i, dev;
	struct tty_struct * tty;

	dev = file->f_rdev;
	if (MAJOR(dev) != 4) {
		printk("tty_read: bad pseudo-major nr #%d\n", MAJOR(dev));
		return -EINVAL;
	}
	dev = MINOR(dev);
	tty = TTY_TABLE(dev);
	if (!tty || (tty->flags & (1 << TTY_IO_ERROR)))
		return -EIO;
	if ((inode->i_rdev != 0x0400) && /* don't stop on /dev/console */
	    (tty->pgrp > 0) &&
	    (current->tty == dev) &&
	    (tty->pgrp != current->pgrp))
		if (is_ignored(SIGTTIN) || is_orphaned_pgrp(current->pgrp))
			return -EIO;
		else {
			(void) kill_pg(current->pgrp, SIGTTIN, 1);
			return -ERESTARTSYS;
		}
	i = tty->la.read_chan(tty,file,buf,count);
	if (i > 0)
		inode->i_atime = CURRENT_TIME;
	return i;
}

static int tty_write(struct inode * inode, struct file * file, char * buf, int count)
{
	int dev, i, is_console;
	struct tty_struct * tty;

	dev = file->f_rdev;
	is_console = (inode->i_rdev == 0x0400);
	if (MAJOR(dev) != 4) {
		printk("tty_write: pseudo-major != 4\n");
		return -EINVAL;
	}
	dev = MINOR(dev);
	if (is_console && redirect)
		tty = redirect;
	else
		tty = TTY_TABLE(dev);
	if (!tty || !tty->write || (tty->flags & (1 << TTY_IO_ERROR)))
		return -EIO;
	if (!is_console && L_TOSTOP(tty) && (tty->pgrp > 0) &&
	    (current->tty == dev) && (tty->pgrp != current->pgrp)) {
		if (is_orphaned_pgrp(current->pgrp))
			return -EIO;
		if (!is_ignored(SIGTTOU)) {
			(void) kill_pg(current->pgrp, SIGTTOU, 1);
			return -ERESTARTSYS;
		}
	}
	i = tty->la.write_chan(tty,file,buf,count);
	if (i > 0)
		inode->i_mtime = CURRENT_TIME;
	return i;
}


/*
 * This is so ripe with races that you should *really* not touch this
 * unless you know exactly what you are doing. All the changes have to be
 * made atomically, or there may be incorrect pointers all over the place.
 */
/* added by KA, temporary */
static struct tty_struct *
get_tty_struct() {
	unsigned short *buf1, *buf2, *buf3;
	struct tty_struct *tty;

	tty = (struct tty_struct *) get_free_page(GFP_KERNEL);
	if (tty == NULL)
		return NULL;
	memset(tty, 0, 4096);
	buf1 = kmalloc(QUEUE_BLOCK_SIZE, GFP_KERNEL);
	buf2 = kmalloc(QUEUE_BLOCK_SIZE, GFP_KERNEL);
	buf3 = kmalloc(QUEUE_BLOCK_SIZE, GFP_KERNEL);
	if (buf1 == NULL || buf2 == NULL || buf3 == NULL) {
		if (buf1)  kfree_s(buf1, QUEUE_BLOCK_SIZE);
		if (buf2)  kfree_s(buf2, QUEUE_BLOCK_SIZE);
		if (buf3)  kfree_s(buf3, QUEUE_BLOCK_SIZE);
		free_page((unsigned long) tty);
		return NULL;
	}
	tty->read_q.buf = buf1;
	tty->write_q.buf = buf2;
	tty->secondary.buf = buf3;
	return tty;
}

static void
free_tty_struct(struct tty_struct *tty) {
	kfree_s(tty->read_q.buf, QUEUE_BLOCK_SIZE);
	kfree_s(tty->write_q.buf, QUEUE_BLOCK_SIZE);
	kfree_s(tty->secondary.buf, QUEUE_BLOCK_SIZE);
	free_page((unsigned long) tty);
}

static int init_dev(int dev)
{
	struct tty_struct *tty, *o_tty;
	struct termios *tp, *o_tp;
	int retval;
	int o_dev;

	o_dev = PTY_OTHER(dev);
	tty = o_tty = NULL;
	tp = o_tp = NULL;
repeat:
	retval = -EAGAIN;
	if (IS_A_PTY_MASTER(dev) && tty_table[dev] && tty_table[dev]->count)
		goto end_init;
	retval = -ENOMEM;
	if (!tty_table[dev] && !tty) {
		tty = get_tty_struct();
		if (!tty)
			goto end_init;
		initialize_tty_struct(dev, tty);
		goto repeat;
	}
	if (!tty_termios[dev] && !tp) {
		tp = (struct termios *) kmalloc(sizeof(struct termios), GFP_KERNEL);
		if (!tp)
			goto end_init;
		initialize_termios(dev, tp);
		goto repeat;
	}
	if (IS_A_PTY(dev)) {
		if (!tty_table[o_dev] && !o_tty) {
			o_tty = get_tty_struct();
			if (!o_tty)
				goto end_init;
			initialize_tty_struct(o_dev, o_tty);
			goto repeat;
		}
		if (!tty_termios[o_dev] && !o_tp) {
			o_tp = (struct termios *) kmalloc(sizeof(struct termios), GFP_KERNEL);
			if (!o_tp)
				goto end_init;
			initialize_termios(o_dev, o_tp);
			goto repeat;
		}
	}
	/* Now we have allocated all the structures: update all the pointers.. */
	if (!tty_termios[dev]) {
		tty_termios[dev] = tp;
		tp = NULL;
	}
	if (!tty_table[dev]) {
		tty->termios = tty_termios[dev];
		SOFT_INT_OFF;
		tty_table[dev] = tty;
		tty->next = tty_list;
		tty_list = tty;
		if (ldiscs[0]->open_ldisc)
			ldiscs[0]->open_ldisc(tty);
		SOFT_INT_ON;
		tty = NULL;
	}
	if (IS_A_PTY(dev)) {
		if (!tty_termios[o_dev]) {
			tty_termios[o_dev] = o_tp;
			o_tp = NULL;
		}
		if (!tty_table[o_dev]) {
			o_tty->termios = tty_termios[o_dev];
			tty_table[o_dev] = o_tty;
			o_tty->next = tty_list;
			tty_list = o_tty;
			o_tty = NULL;
		}
		tty_table[dev]->link = tty_table[o_dev];
		tty_table[o_dev]->link = tty_table[dev];
	}
	tty_table[dev]->count++;
	if (IS_A_PTY_MASTER(dev))
		tty_table[o_dev]->count++;
	retval = 0;
end_init:
	if (tty)
		free_tty_struct(tty);
	if (o_tty)
		free_tty_struct(o_tty);
	if (tp)
		kfree_s(tp, sizeof(struct termios));
	if (o_tp)
		kfree_s(o_tp, sizeof(struct termios));
	return retval;
}

/*
 * Even releasing the tty structures is a tricky business.. We have
 * to be very careful that the structures are all released at the
 * same time, as interrupts might otherwise get the wrong pointers.
 */
static void remove_from_tty_list(struct tty_struct *tty)
{
	struct tty_struct *ttyp, **ttypp;

	ttypp = &tty_list;
	for (;;) {
		ttyp = *ttypp;
		if (ttyp == NULL) {
			printk("tty 0x%x not on tty list\n", tty);
			break;
		}
		if (ttyp == tty) {
			*ttypp = ttyp->next;
			break;
		}
		ttypp = &ttyp->next;
	}
	for (ttyp = tty_list ; ttyp ; ttyp = ttyp->next) {	/*DEBUG*/
		if (ttyp == tty)
			printk("tty 0x%x still on tty list\n", tty);
	}
}

static void release_dev(int dev, struct file * filp)
{
	struct tty_struct *tty, *o_tty;
#ifdef notdef
	struct tty_struct *ttyp, **ttypp;   /* for deletion from tty_list */
#endif
	struct termios *tp, *o_tp;

	TTY_TRACE(("release_dev %d called:  ", dev));
	tty = tty_table[dev];
	tp = tty_termios[dev];
	o_tty = NULL;
	o_tp = NULL;
	if (!tty) {
		printk("release_dev: tty_table[%d] was NULL\n", dev);
		return;
	}
	if (!tp) {
		printk("release_dev: tty_termios[%d] was NULL\n", dev);
		return;
	}
	if (IS_A_PTY(dev)) {
		TTY_TRACE(("pty, "));
		o_tty = tty_table[PTY_OTHER(dev)];
		o_tp = tty_termios[PTY_OTHER(dev)];
		if (!o_tty) {
			printk("release_dev: pty pair(%d) was NULL\n", dev);
			return;
		}
		if (!o_tp) {
			printk("release_dev: pty pair(%d) termios was NULL\n", dev);
			return;
		}
		if (tty->link != o_tty || o_tty->link != tty) {
			printk("release_dev: bad pty pointers\n");
			return;
		}
	}
	tty->write_data_cnt = 0; /* Clear out pending trash */
	TTY_TRACE(("calling device close, "));
	if (tty->close)
		tty->close(tty, filp);
	if (IS_A_PTY_MASTER(dev)) {
		TTY_TRACE(("pty master, "));
		if (--tty->link->count < 0) {
			printk("release_dev: bad tty slave count (dev = %d): %d\n",
			       dev, tty->count);
			tty->link->count = 0;
		}
	}
	if (--tty->count < 0) {
		printk("release_dev: bad tty_table[%d]->count: %d\n",
		       dev, tty->count);
		tty->count = 0;
	}
	TTY_TRACE(("tty->count=%d\n", tty->count));
	if (tty->count)
		return;
	TTY_TRACE(("Closing line discipline, "));
	if (tty->la.close_ldisc)
		tty->la.close_ldisc(tty);
	if (o_tty) {
		if (o_tty->count) {
			TTY_TRACE(("o_tty->count=%d\n", o_tty->count));
			return;
		} else {
			remove_from_tty_list(o_tty);
			tty_table[PTY_OTHER(dev)] = NULL;
			tty_termios[PTY_OTHER(dev)] = NULL;
		}
	}
#ifdef notdef
	ttypp = &tty_list;
	for (;;) {
		ttyp = *ttypp;
		if (ttyp == NULL) {
			printk("tty 0x%x not on tty list\n", tty);
			break;
		}
		if (ttyp == tty) {
			*ttypp = ttyp->next;
			break;
		}
		ttypp = &ttyp->next;
	}
	for (ttyp = tty_list ; ttyp ; ttyp = ttyp->next) {	/*DEBUG*/
		if (ttyp == tty)
			printk("tty 0x%x still on tty list\n", tty);
	}
#else
	remove_from_tty_list(tty);
#endif
	tty_table[dev] = NULL;
	if (IS_A_PTY(dev)) {
		TTY_TRACE(("pty, "));
		tty_termios[dev] = NULL;
		kfree_s(tp, sizeof(struct termios));
	}
	if (tty == redirect || o_tty == redirect)
		redirect = NULL;
	TTY_TRACE(("free_page 0x%X, ", tty));
	free_tty_struct(tty);
	if (o_tty) {
		TTY_TRACE(("free o_tty 0x%X, ", o_tty));
		free_tty_struct(o_tty);
	}
	if (o_tp)
		kfree_s(o_tp, sizeof(struct termios));
	TTY_TRACE(("done\n"));
}

/*
 * tty_open and tty_release keep up the tty count that contains the
 * number of opens done on a tty. We cannot use the inode-count, as
 * different inodes might point to the same tty.
 *
 * Open-counting is needed for pty masters, as well as for keeping
 * track of serial lines: DTR is dropped when the last close happens.
 * (This is not done solely through tty->count, now.  - Ted 1/27/92)
 *
 * The termios state of a pty is reset on first open so that
 * settings don't persist across reuse.
 */
static int tty_open(struct inode * inode, struct file * filp)
{
	struct tty_struct *tty;
	int major, minor;
	int noctty, retval;
	static int first = 1;

	TTY_TRACE(("tty_open %d, %d: ",
			MAJOR(inode->i_rdev), MINOR(inode->i_rdev)));
	if (first) {
		printk("Initializing tty timer\n");
		bh_base[TTYTIMER_BH].routine = tty_timer;
		tty_trigger_timer();
		first = 0;
	}
	minor = MINOR(inode->i_rdev);
	major = MAJOR(inode->i_rdev);
	noctty = filp->f_flags & O_NOCTTY;
	if (major == 5) {
		if (!minor) {
			major = 4;
			minor = current->tty;
		}
		noctty = 1;
	} else if (major == 4) {
		if (!minor) {
			minor = fg_console + 1;
			noctty = 1;
		}
	} else {
		printk("Bad major #%d in tty_open\n", MAJOR(inode->i_rdev));
		return -ENODEV;
	}
	if (minor <= 0)
		return -ENXIO;
	if (IS_A_PTY_MASTER(minor))
		noctty = 1;
	filp->f_rdev = (major << 8) | minor;
	retval = init_dev(minor);
	if (retval)
		return retval;
	tty = tty_table[minor];

	/* clean up the packet stuff. */
	/*
	 *  Why is this not done in init_dev?  Right here, if another 
	 * process opens up a tty in packet mode, all the packet 
	 * variables get cleared.  Come to think of it, is anything 
	 * using the packet mode at all???  - Ted, 1/27/93
	 */
	tty->status_changed = 0;
	tty->ctrl_status = 0;
	tty->packet = 0;

	if (tty->open) {
		retval = tty->open(tty, filp);
	} else {
		retval = -ENODEV;
	}
	if (retval) {
		release_dev(minor, filp);
		TTY_TRACE(("open failed %d\n", retval));
		return retval;
	}
	if (!noctty &&
	    current->leader &&
	    current->tty<0 &&
	    tty->session==0) {
		current->tty = minor;
		tty->session = current->session;
		tty->pgrp = current->pgrp;
	}
	filp->f_rdev = 0x0400 | minor; /* Set it to something normal */
	TTY_TRACE(("open succeeded\n"));
	return 0;
}

/*
 * Note that releasing a pty master also releases the child, so
 * we have to make the redirection checks after that and on both
 * sides of a pty.
 */
static void tty_release(struct inode * inode, struct file * filp)
{
	int dev;

	TTY_TRACE(("tty_release called\n"));
	dev = filp->f_rdev;
	if (MAJOR(dev) != 4) {
		printk("tty_release: tty pseudo-major != 4\n");
		return;
	}
	dev = MINOR(filp->f_rdev);
	if (!dev) {
		printk("tty_release: bad f_rdev\n");
		return;
	}
	release_dev(dev, filp);
}


static int tty_select(struct inode * inode, struct file * filp, int sel_type, select_table * wait)
{
	struct tty_struct * tty = TTY_TABLE(MINOR(filp->f_rdev));

	if (!tty) {
		printk("tty_select: tty struct for dev %d was NULL\n",
			filp->f_rdev);
		return 0;
	}
	return tty->la.select(inode, filp, sel_type, wait);
}

/*
 * This implements the "Secure Attention Key" ---  the idea is to
 * prevent trojan horses by killing all processes associated with this
 * tty when the user hits the "Secure Attention Key".  Required for
 * super-paranoid applications --- see the Orange Book for more details.
 * 
 * This code could be nicer; ideally it should send a HUP, wait a few
 * seconds, then send a INT, and then a KILL signal.  But you then
 * have to coordinate with the init process, since all processes associated
 * with the current tty must be dead before the new getty is allowed
 * to spawn.
 */
void do_SAK( struct tty_struct *tty)
{
	struct task_struct **p;
	int line = tty->line;
	int session = tty->session;
	int		i;
	struct file	*filp;
	
	tty->la.flush_input(tty);
	tty->la.flush_output(tty);
 	for (p = &LAST_TASK ; p > &FIRST_TASK ; --p) {
		if (!(*p))
			continue;
		if (((*p)->tty == line) ||
		    ((session > 0) && ((*p)->session == session)))
			send_sig(SIGKILL, *p, 1);
		else {
			for (i=0; i < NR_FILE; i++) {
				filp = (*p)->filp[i];
				if (filp && (filp->f_op == &tty_fops) &&
				    (MINOR(filp->f_rdev) == line)) {
					send_sig(SIGKILL, *p, 1);
					break;
				}
			}
		}
	}
}


/*
 * This routine allows a kernel routine to send a large chunk of data
 * to a particular tty; if all of the data can be queued up for ouput
 * immediately, tty_write_data() will return 0.  If, however, not all
 * of the data can be immediately queued for delivery, the number of
 * bytes left to be queued up will be returned, and the rest of the
 * data will be queued up when there is room.  The callback function
 * will be called (with the argument callarg) when the last of the
 * data is finally in the queue.
 *
 * Note that the callback routine will _not_ be called if all of the
 * data could be queued immediately.  This is to avoid a problem with
 * the kernel stack getting too deep, which might happen if the
 * callback routine calls tty_write_data with itself as an argument.
 */
int tty_write_data(struct tty_struct *tty, char *bufp, int buflen,
		    void (*callback)(void * data), void * callarg)
{
	int head, tail, count;
	unsigned long flags;
	char *p;

#define VLEFT ((tail-head-1)&(TTY_BUF_SIZE-1))

	__asm__ __volatile__("pushfl ; popl %0 ; cli":"=r" (flags));
	if (tty->write_data_cnt) {
		__asm__ __volatile__("pushl %0 ; popfl"::"r" (flags));
		return -EBUSY;
	}

	head = tty->write_q.head;
	tail = tty->write_q.tail;
	count = buflen;
	p = bufp;

	while (count && VLEFT > 0) {
		tty->write_q.buf[head++] = *p++;
		head &= TTY_BUF_SIZE-1;
	}
	tty->write_q.head = head;
	if (count) {
		tty->write_data_cnt = count;
		tty->write_data_ptr = p;
		tty->write_data_callback = callback;
		tty->write_data_arg = callarg;
	}
	__asm__ __volatile__("pushl %0 ; popfl"::"r" (flags));
	return count;
}

/*
 * This routine routine is called after an interrupt has drained a
 * tty's write queue, so that there is more space for data waiting to
 * be sent in tty->write_data_ptr.
 *
 * tty_check_write[8] is a bitstring which indicates which ttys
 * needs to be processed.
 */
void tty_bh_routine()
{
	int	i, j, line, mask;
	int	head, tail, count;
	unsigned char * p;
	struct tty_struct * tty;

	for (i = 0, line = 0; i < MAX_TTYS / 32; i++) {
		if (!tty_check_write[i]) {
			line += 32;
			continue;
		}
		for (j=0, mask=0; j < 32; j++, line++, mask <<= 1) {
			if (!clear_bit(j, &tty_check_write[i])) {
				tty = tty_table[line];
				if (!tty || !tty->write_data_cnt)
					continue;
				cli();
				head = tty->write_q.head;
				tail = tty->write_q.tail;
				count = tty->write_data_cnt;
				p = tty->write_data_ptr;

				while (count && VLEFT > 0) {
					tty->write_q.buf[head++] = *p++;
					head &= TTY_BUF_SIZE-1;
				}
				tty->write_q.head = head;
				tty->write_data_ptr = p;
				tty->write_data_cnt = count;
				sti();
				if (!count)
					(tty->write_data_callback)
						(tty->write_data_arg);
			}
		}
	}
	
}

/*
 * This subroutine initializes a tty structure.  We have to set up
 * things correctly for each different type of tty.
 */
static void initialize_tty_struct(int line, struct tty_struct *tty)
{
	tty->line = line;
	tty->pgrp = -1;
	tty->winsize.ws_row = 24;
	tty->winsize.ws_col = 80;

	/* set up line discipline. */
	tty->la = ldiscs[0]->ldisc_above;
	tty->lb = ldiscs[0]->ldisc_below;

	if (IS_A_CONSOLE(line)) {
		tty->open = con_open;
		tty->winsize.ws_row = video_num_lines;
		tty->winsize.ws_col = video_num_columns;
	} else if IS_A_SERIAL(line) {
		tty->open = rs_open;
	} else if IS_A_PTY(line) {
		tty->open = pty_open;
	}
}

static void initialize_termios(int line, struct termios * tp)
{
	memset(tp, 0, sizeof(struct termios));
	memcpy(tp->c_cc, INIT_C_CC, NCCS);
	tp->c_line = ldiscs[0]->name;
	if (IS_A_CONSOLE(line)) {
		tp->c_iflag = ICRNL | IXON;
		tp->c_oflag = OPOST | ONLCR;
		tp->c_cflag = B38400 | CS8 | CREAD;
		tp->c_lflag = ISIG | ICANON | ECHO |
			ECHOCTL | ECHOKE;
	} else if (IS_A_SERIAL(line)) {
		tp->c_cflag = B2400 | CS8 | CREAD | HUPCL | CLOCAL;
	} else if (IS_A_PTY_MASTER(line)) {
		tp->c_cflag = B9600 | CS8 | CREAD;
	} else if (IS_A_PTY_SLAVE(line)) {
		tp->c_iflag = ICRNL | IXON;
		tp->c_oflag = OPOST | ONLCR;
		tp->c_cflag = B38400 | CS8 | CREAD;
		tp->c_lflag = ISIG | ICANON | ECHO |
			ECHOCTL | ECHOKE;
	}
}

long tty_init(long kmem_start)
{
	int i;
	extern struct file_operations * chrdev_fops[];

	if (sizeof(struct tty_struct) > 4096)
		panic("size of tty structure > 4096!");
#if 0 /* register_chrdev does not exist prior to 0.99.pl6 */
	if (register_chrdev(4,"tty",&tty_fops))
		panic("unable to get major 4 for tty device");
	if (register_chrdev(5,"tty",&tty_fops))
		panic("unable to get major 5 for tty device");
#else
	chrdev_fops[4] = &tty_fops;
	chrdev_fops[5] = &tty_fops;
#endif
	for (i = 0 ; i < MAX_TTYS; i++) {
		tty_table[i] =  0;
		tty_termios[i] = 0;
	}
	memset(tty_check_write, 0, sizeof(tty_check_write));
	bh_base[TTY_BH].routine = tty_bh_routine;

	kmem_start = kbd_init(kmem_start);
	kmem_start = con_init(kmem_start);
	kmem_start = rs_init(kmem_start);
	/* printk("%d virtual consoles\n\r",NR_CONSOLES); */
	return kmem_start;
}


void flush_input(struct tty_struct *tty)
{
	tty->la.flush_input(tty);
}

void flush_output(struct tty_struct *tty)
{
	tty->la.flush_output(tty);
}


void tty_no_op(struct tty_struct *tty)
{
	/* nothing to do for this function */
}
